import configparser
import logging
import os
import subprocess
import sys
import tempfile
import vcf as pyvcf
import pysam
import inspect
from glob import glob

def tabix(f):
    """
    Generate a tabix index for a bgzipped file

    :param f: path to a bgzip compressed file
    :type f: str

    :returns: Nothing
    :rtype: None
    """

    tabix_call = "tabix -f " + f
    logging.debug(tabix_call)
    subprocess.run(tabix_call, shell=True)

def check_missing_file(file_list, die=True):
    missing_files = []
    for item in file_list:
        if not os.path.isfile(item):
            missing_files.append(item)
    if die and len(missing_files) > 0:
        raise ValueError("Missing these files: " + ",".join(missing_files))
    return missing_files

def tmp_mity_file_name():
    """
    Create a tmp mity vcf file

    #TODO There must be a more pythonic way of doing this.
    """
    f = tempfile.NamedTemporaryFile(mode="wt", prefix='mity', suffix=".vcf",
                                    delete=False)
    f.close()
    assert isinstance(f.name, str)
    return f.name


def create_prefix(file_name, prefix=None):
    """
    Most mity functions have an optional prefix. If a prefix is not specified,
    then use the  file name (minus the .vcf.gz or .bam suffix) as the prefix.
    :param file_name: The vcf, bam, bed filename
    :param prefix: The optional prefix. If None, then create a prefix from
    file_name, else return prefix
    :return: str prefix
    """
    if prefix is not None:
        pass
    elif ".vcf" in file_name:
        prefix = [os.path.basename(file_name).split(".vcf")[0], prefix][prefix is not None]
    elif ".bam" in file_name:
        prefix = [os.path.basename(file_name).split(".bam")[0], prefix][prefix is not None]
    else:
        raise ValueError("Unsupported file type")
    return prefix


def write_vcf(new_vcf, out_file, genome_file='mitylib/reference/b37d5.genome'):
    """
    write a vcf object to vcf.gz file with tbi index.

    This differs from write_merged_vcf, as mity merge doesn't split each VCF
    line on tabs, whereas mity normalise does

    :param new_vcf: new_vcf is a list of lists, created by normalise
    :param out_file: the resulting filename. this should end in vcf.gz
    :return: None. This function writes a vcf.gz and vcf.gz.tbi file.
    """
    f = tmp_mity_file_name()
    logging.debug("Writing uncompressed vcf to " + f)
    with open(f, mode='wt', encoding='utf-8') as myfile:
        for vcf_line in new_vcf:
            myfile.write('\t'.join([str(elem) for elem in vcf_line]) + '\n')
    gsort_vcf(f, out_file, genome_file=genome_file)
    # bcftools_sort_vcf(f, out_file)


def gsort_vcf(f, out_file, genome_file='mitylib/reference/b37d5.genome', remove_unsorted_vcf=False):
    """
    use gsort to sort the records in a VCF file according to a .genome file.

    :param f: the path to an unsorted vcf.gz file
    :param out_file: the path to a resulting sorted vcf.gz file
    :param genome_file: the .genome file corresponding to the reference genome. see https://github.com/brentp/gsort
    :param remove_unsorted_vcf: if True, then the input file 'f' will be deleted.
    :return: nothing
    """
    logging.debug("Sorting, bgzipping {} -> {}".format(f, out_file))
    logging.debug("gsort is using genome file " + genome_file)
    gsort_cmd = "gsort {} {} | bgzip -cf > {}".format(f, genome_file, out_file)

    logging.debug(gsort_cmd)

    subprocess.run(gsort_cmd, shell=True)
    logging.debug("Tabix indexing {}".format(out_file))
    tabix(out_file)
    if remove_unsorted_vcf:
        os.remove(f)

def bcftools_sort_vcf(f, out_file, remove_unsorted_vcf=False):
    """
    use bcftools sort to sort the records in a VCF file.

    :param f: the path to an unsorted vcf.gz file
    :param out_file: the path to a resulting sorted vcf.gz file
    :param remove_unsorted_vcf: if True, then the input file 'f' will be deleted.
    :return: nothing
    """
    logging.debug("Sorting, bgzipping {} -> {}".format(f, out_file))
    bcftools_sort_cmd = "bcftools sort {} -O z -o {} 2>&1 >/dev/null".format(f, out_file)

    logging.debug(bcftools_sort_cmd)

    subprocess.run(bcftools_sort_cmd, shell=True)
    logging.debug("Tabix indexing {}".format(out_file))
    tabix(out_file)
    if remove_unsorted_vcf:
        os.remove(f)

def write_merged_vcf(new_vcf, out_file, genome_file='mitylib/reference/b37d5.genome'):
    """
    write a vcf object to vcf.gz file with tbi index.

    This differs from write_vcf, as mity merge doesn't split each VCF line on
    tabs, whereas mity normalise does

    :param new_vcf: new_vcf is a list of strings, created by merge
    :param out_file: the resulting filename. this should end in vcf.gz
    :return: None. This function writes a vcf.gz and vcf.gz.tbi file.
    """
    f = tmp_mity_file_name()
    logging.debug("Writing uncompressed vcf to " + f)
    with open(f, mode='wt', encoding='utf-8') as myfile:
        for vcf_line in new_vcf:
            myfile.write(vcf_line + '\n')
    gsort_vcf(f, out_file, genome_file=genome_file)

def create_genome_file(vcf_file, genome_file):
    """
    gsort (https://github.com/brentp/gsort) requires a '.genome'
    file to tell it how to sort the vcf records. This function creates a
    '.genome' file in the same order as the contig lines in the vcf header.

    :param vcf_file: a vcf file with the correct contig names
    :param genome_file: the resulting .genome file
    :return: None. this creates a '.genome' file
    """

    vcf = pyvcf.Reader(filename=vcf_file)
    with open(genome_file, mode='wt', encoding='utf-8') as genome_file:
        for contig in vcf.contigs.keys():
            genome_file.write('\t'.join(vcf.contigs[contig]))

def check_dependency(dep, exit=True):
    """
    Check if a dependency exists.

    Special cases:
    * gsort: this utility from brentp has the same name as GNU sort on some linux
      systems. This function will check that brentp's gsort is found.

    >>> check_dependency("ls")
    >>> check_dependency("freebayes")
    >>> check_dependency("gsort")
    :param dep: name of the dependency
    :param exit: If True, then if the dependency isn't found, the session will exit.
    :return: True/False if dependency was found.
    """
    found = False
    try:
        res = subprocess.run(['which', dep], capture_output=True, check=True, encoding="UTF8")
        found = True
        logging.info("Found dependency: " + str(res.stdout).strip())
        if dep == "gsort":
            # There is a potential name clash with gnu sort on linux
            res = subprocess.check_output('gsort --help | grep -c GENOME', shell=True)
            if res != b'2\n':
                logging.error("Adjust your PATH to ensure that brentp's gsort is found before GNU sort" + dep)
    except subprocess.CalledProcessError:
        logging.error("Missing dependency: " + dep)
        if exit:
            sys.exit(1)
    return found


def check_dependencies(f='verchew.ini'):
    """
    Check that all of the dependencies that are listed in the INI format exist.
    :param f: an INI formatted file. see verchew [https://verchew.readthedocs.io/en/latest/]
    :return:
    """
    config = configparser.ConfigParser()
    config.read(f)
    for section in config.sections():
        dependency = config[section]['cli']
        check_dependency(dependency)


def make_hgvs(pos, ref, alt):
    # make HGVS syntax
    if len(alt) > 1 or len(ref) > 1:
        # this is an indel
        if len(ref) > len(alt):
            # this is a del
            delet = ref[1:]
            if len(delet) == 1:
                hgvs_pos = int(pos) + 1
                # print(hgvs_pos)
            elif len(delet) > 1:
                hvgs_pos_start = int(pos) + 1
                hvgs_pos_end = int(pos) + len(delet)
                hgvs_pos = str(hvgs_pos_start) + "_" + str(hvgs_pos_end)
                # print(hgvs_pos)
            # print(hgvs_pos)
            hgvs = "m." + str(hgvs_pos) + "del"
            # print(hgvs)

        else:
            # print("ins")
            # this is an ins
            ins = alt[1:]
            if len(ins) == 1:
                hgvs_pos = int(pos) + 1
                # print(hgvs_pos)
            elif len(ins) > 1:
                hvgs_pos_start = int(pos) + 1
                hvgs_pos_end = int(pos) + len(ins)
                hgvs_pos = str(hvgs_pos_start) + "_" + str(hvgs_pos_end)
                # print(hgvs_pos)
            # print(hgvs_pos)
            hgvs = "m." + str(hgvs_pos) + "ins"
            # print(hgvs)

    else:
        # this is a SNP
        hgvs = "m." + str(pos) + str(ref) + ">" + str(alt)
    return hgvs

def select_reference_fasta(reference, custom_reference_fa=None):
    """
    Allow the user to select one of the pre-loaded reference genome fasta files, via --reference,
    or supply their own via --custom_reference. This function will return the path to
    the reference genome fasta file.

    :param reference: one of the inbuilt reference genomes: hs37d5, hg19, hg38, mm10.
    :param custom_reference_fa: the path to a custom reference genome, or None. If this
    file exists, then it will override the option provided by 'reference'.
    :return the path to the reference genome as a str.

    >>> select_reference_fasta('hg19', None)
    'reference/hg19.chrM.fa'
    >>> select_reference_fasta('hg19', 'mitylib/reference/hs37d5.MT.fa')
    'reference/hs37d5.MT.fa'
    >>> select_reference_fasta('hg19', 'nonexistent.fa')
    'reference/hg19.chrM.fa'

    """
    if custom_reference_fa is not None and os.path.exists(custom_reference_fa):
        res = custom_reference_fa
    else:
        ref_dir = os.path.join(get_mity_dir(), 'reference')
        res = glob('{}/{}.*.fa'.format(ref_dir, reference))
        logging.debug(",".join(res))
        assert len(res) == 1
        res = res[0]
    return res

def select_reference_genome(reference, custom_reference_genome=None):
    """
    This function returns the path to a '.genome' file [1], which is needed for `gsort` to order the chromosomes
    properly.

    [1]: https://bedtools.readthedocs.io/en/latest/content/general-usage.html#genome-file-format

    :param reference: one of the inbuilt reference genomes: hs37d5, hg19, hg38, mm10
    :param custom_reference_genome: the path to a custom reference .genome file, or None. If this
    file exists, then it will override the option provided by 'reference'.
    :return the path to the reference .genome file as a str.

    >>> select_reference_genome('hg19', None)
    'reference/hg19.genome'
    >>> select_reference_genome('hg19', 'mitylib/reference/hs37d5.MT.fa')
    'reference/hs37d5.genome'
    >>> select_reference_genome('hg19', 'nonexistent.fa')
    'reference/hg19.genome'

    """
    if custom_reference_genome is not None and os.path.exists(custom_reference_genome):
        res = custom_reference_genome
    else:
        ref_dir = os.path.join(get_mity_dir(), 'reference')
        logging.debug("Looking for .genome file in " + ref_dir)
        res = glob('{}/{}.genome'.format(ref_dir, reference))
        logging.debug(",".join(res))
        assert len(res) == 1
        res = res[0]
    return res

def get_mity_dir():
    path = os.path.dirname(sys.modules['mitylib'].__file__)
    return path

def vcf_get_mt_contig(vcf):
    """
    get the mitochondrial contig name and length from a VCF file
    :param vcf: path to a vcf file
    :return: a tuple of contig name as str and length as int

    >>> vcf_get_mt_contig('./151016_FR07959656.dedup.realigned.recalibrated.chrMT.dedup.realigned.recalibrated.chrMT.mity.vcf.gz')
    ('MT', 16569)
    """
    r = pyvcf.Reader(filename=vcf, compressed=True)
    chroms = r.contigs.keys()
    mito_contig = {'MT', 'chrM'}.intersection(chroms)
    assert len(mito_contig) == 1
    mito_contig = ''.join(mito_contig)

    return r.contigs[mito_contig].id, r.contigs[mito_contig].length


def bam_get_mt_contig(bam, as_string=False):
    """
    get the mitochondrial contig name and length from a BAM file
    :param bam: path to a bam file
    :return: a tuple of contig name as str and length as int

    >>> bam_get_mt_contig('NA12878.alt_bwamem_GRCh38DH.20150718.CEU.low_coverage.chrM.bam', False)
    ('chrM', 16569)
    >>> bam_get_mt_contig('NA12878.alt_bwamem_GRCh38DH.20150718.CEU.low_coverage.chrM.bam', True)
    'chrM:1-16569'
    """
    r = pysam.AlignmentFile(bam, "rb")
    chroms = [str(record.get("SN")) for record in r.header['SQ']]
    mito_contig = {'MT', 'chrM'}.intersection(chroms)
    assert len(mito_contig) == 1
    mito_contig = ''.join(mito_contig)
    res = None
    for record in r.header['SQ']:
        if mito_contig == record['SN']:
            res = record['SN'], record['LN']
    if res is not None and as_string:
        res = res[0] + ":1-" + str(res[1])
    return res

def bam_has_RG(bam):
    """
    Does the BAM File have an @RG header? This is critical for mity to correctly call variants.

    :param bam: str: path to bam file
    :return: True/False
    >>> bam_has_RG('NA12878.alt_bwamem_GRCh38DH.20150718.CEU.low_coverage.chrM.bam')
    """
    r = pysam.AlignmentFile(bam, "rb")
    return len(r.header['RG']) > 0

def get_annot_file(f):
    #mitylibdir = os.path.dirname(inspect.getfile(mitylib))
    #mitylibdir = os.path.dirname(mitylib.__file__)
    mitylibdir = get_mity_dir()
    p = os.path.join(mitylibdir, "annot", f)
    assert os.path.exists(p)
    return p



"""
Mity: a sensitive variant analysis pipeline optimised for WGS data



Usage: See the online manual for details: http://github.com/KCCG/mity
Authors: Clare Puttick, Mark Cowley
License: MIT
"""
import argparse
import logging

from . import (call, normalise, report, merge)
from ._version import __version__
from .util import select_reference_fasta
from .util import select_reference_genome

__all__ = []


def public(fn):
    __all__.append(fn.__name__)
    return fn


usage = __doc__.split('\n\n\n', maxsplit=1)
usage[-1] += "Version: " + __version__

AP = argparse.ArgumentParser(description=usage[0], epilog=usage[1], formatter_class=argparse.RawTextHelpFormatter)
AP_subparsers = AP.add_subparsers(
        help="mity sub-commands (use with -h for more info)")

# call -------------------------------------------------------------------------

do_call = public(call.do_call)

def _cmd_call(args):
    if True:
        # TODO: why does this not turn on debugging output? It needs to be turned on in `mity:L20` to work
        logging.basicConfig(level=logging.DEBUG, format="%(message)s")
        #logging.getLogger('mitylib').setLevel(level=logging.DEBUG)

    """Call mitochondrial variants"""
    logging.info("mity version %s", __version__)
    logging.info("Calling mitochondrial variants")
    logging.debug("Debugging mode activated")

    genome = select_reference_genome(args.reference, None)
    args.reference = select_reference_fasta(args.reference, None)

    call.do_call(args.bam, args.reference, genome, args.prefix, args.min_mq,
                 args.min_bq, args.min_af, args.min_ac, args.p, args.normalise,
                 args.out_folder_path, args.region)


P_call = AP_subparsers.add_parser('call', help=_cmd_call.__doc__)
P_call.add_argument('bam', action='append', nargs='+',
                    help='BAM files to run the analysis on.')
P_call.add_argument('--reference', choices=['hs37d5', 'hg19', 'hg38', 'mm10'],
                    default="hs37d5", required=False,
                    help='reference genome version to use. default: hs37d5')
# P_call.add_argument('--custom_reference', action='store',
#                     default="", required=False,
#                     help='The path to a custom reference genome file in uncompressed fasta format')
P_call.add_argument('--prefix', action='store',
                    help='Output files will be named with PREFIX')
P_call.add_argument('--min-mapping-quality', action='store', type=int,
                    default=30,
                    help='Exclude alignments from analysis if they have a '
                         'mapping quality less than MIN_MAPPING_QUALITY. '
                         'Default: 30',
                    dest="min_mq")
P_call.add_argument('--min-base-quality', action='store', type=int, default=24,
                    help='Exclude alleles from analysis if their supporting '
                         'base quality is less than MIN_BASE_QUALITY. '
                         'Default: 24',
                    dest="min_bq")
P_call.add_argument('--min-alternate-fraction', action='store', type=float,
                    default=0.01,
                    help='Require at least MIN_ALTERNATE_FRACTION '
                         'observations supporting an alternate allele within '
                         'a single individual in the in order to evaluate the '
                         'position. Default: 0.01, range = [0,1]',
                    dest="min_af")
P_call.add_argument('--min-alternate-count', action='store', type=int,
                    default=4,
                    help='Require at least MIN_ALTERNATE_COUNT observations '
                         'supporting an alternate allele within a single '
                         'individual in order to evaluate the position. '
                         'Default: 4',
                    dest="min_ac")
P_call.add_argument('--p', action='store', type=float,
                    default=0.002,
                    help='Minimum noise level. This is used to calculate QUAL score. '
                         'Default: 0.002, range = [0,1]',
                    dest="p")
P_call.add_argument('--normalise', action='store_true',
                    help='Normalise the resulting VCF?')
P_call.add_argument('--out-folder-path', action='store', type=str,
                    default='.',
                    help='Output files will be saved in OUT_FOLDER_PATH. '
                         "Default: '.' ",
                    dest="out_folder_path")
P_call.add_argument('--region', action='store', type=str,
                    default=None,
                    help='Region of MT genome to call variants in. '
                         'If unset will call variants in entire MT genome as specified in BAM header. '
                         "Default: Entire MT genome. ",
                    dest="region")
P_call.add_argument('--debug', action='store_true',
                    help='Verbose output for debugging?')
P_call.add_argument('--bam-file-list', action='store', type=str,
                    default=None,
                    help='A text file of BAM files to be processed. The path to each file should be on one row per Region of MT genome to call variants in. '
                         'If unset will call variants in entire MT genome as specified in BAM header. '
                         "Default: Entire MT genome. ",
                    dest="region")
P_call.set_defaults(func=_cmd_call)

# normalise --------------------------------------------------------------------

do_normalise = public(normalise.do_normalise)


def _cmd_normalise(args):
    """Normalise & FILTER mitochondrial variants"""
    logging.info("mity %s", __version__)
    logging.info("Normalising and FILTERing mitochondrial vcf.gz file")

    genome = select_reference_genome(args.reference, None)

    normalise.do_normalise(args.vcf, args.outfile, p=args.p, SB_range=[0.1, 0.9], genome=genome)


P_normalise = AP_subparsers.add_parser('normalise', help=_cmd_normalise.__doc__)
P_normalise.add_argument('vcf', action='store',
                         help="vcf.gz file from running mity")
P_normalise.add_argument('--outfile', action='store', required=True,
                         help="output VCF file in bgzip compressed format")
P_normalise.add_argument('--p', action='store', type=float,
                         default=0.002,
                         help='Minimum noise level. This is used to calculate QUAL score'
                              'Default: 0.002, range = [0,1]',
                         dest="p")
P_normalise.add_argument('--reference', choices=['hs37d5', 'hg19', 'hg38', 'mm10'],
                    default="hs37d5", required=False,
                    help='reference genome version to use. default: hs37d5')
P_normalise.set_defaults(func=_cmd_normalise)


# report -----------------------------------------------------------------------

do_report = public(report.do_report)


def _cmd_report(args):
    """Generate mity report"""
    logging.info("mity %s", __version__)
    logging.info("Generating mity report")
    report.do_report(args.vcf, args.prefix, args.min_vaf, args.out_folder_path)


P_report = AP_subparsers.add_parser('report', help=_cmd_report.__doc__)
P_report.add_argument('--prefix', action='store',
                      help='Output files will be named with PREFIX')
P_report.add_argument('--min_vaf', action='store', type=float, default=0, help=
'A variant must have at least this VAF to be included in the report. Default: '
'0.')
P_report.add_argument('--out-folder-path', action='store', type=str,
                    default='.',
                    help='Output files will be saved in OUT_FOLDER_PATH. '
                         "Default: '.' ",
                    dest="out_folder_path")
P_report.add_argument('vcf', action='append', nargs='+',
                    help="mity vcf files to create a report from")
P_report.set_defaults(func=_cmd_report)


# merge -----------------------------------------------------------------------

do_merge = public(merge.do_merge)

def _cmd_merge(args):
    """Merging mity VCF with nuclear VCF"""
    logging.info("mity %s", __version__)
    logging.info("mity vcf merge")

    genome = select_reference_genome(args.reference, None)

    merge.do_merge(args.mity_vcf, args.nuclear_vcf, args.prefix, genome)

P_merge = AP_subparsers.add_parser('merge', help=_cmd_merge.__doc__)
P_merge.add_argument('--mity_vcf', action='store', required=True,
                      help="mity vcf file")
P_merge.add_argument('--nuclear_vcf', action='store', required=True,
                     help="nuclear vcf file")
P_merge.add_argument('--prefix', action='store',
                     help='Output files will be named with PREFIX. '
                     'The default is to use the nuclear vcf name')
P_merge.add_argument('--reference', choices=['hs37d5', 'hg19', 'hg38', 'mm10'],
                     default="hs37d5", required=False,
                     help='reference genome version to use. default: hs37d5')
# P_merge.add_argument('--custom_reference', action='store',
#                      default="", required=False,
#                      help='The path to a custom reference genome file in uncompressed fasta format')
P_merge.set_defaults(func=_cmd_merge)


# version ----------------------------------------------------------------------

def print_version(_args):
    """Display this program's version."""
    print(__version__)


P_version = AP_subparsers.add_parser('version', help=print_version.__doc__)
P_version.set_defaults(func=print_version)


def parse_args(args=None):
    """Parse the command line."""
    return AP.parse_args(args=args)



"""
Run mity-merge.
This merges a nuclear VCF and a mity VCF, such that the MT variants in the
nuclear VCF are replaces with the mity variants, and the headers are merged
"""
import sys
import gzip
import logging
from .util import write_merged_vcf
from .util import create_prefix
from .util import vcf_get_mt_contig

def check_vcf_merge_compatibility(mity_vcf, hc_vcf):
    """
    Check that the MT sequence names and lengths match in both the mity and hc vcf's.
    """
    m = vcf_get_mt_contig(mity_vcf)
    h = vcf_get_mt_contig(hc_vcf)
    return m == h

def do_merge(mity_vcf, hc_vcf, prefix=None, genome='mitylib/reference/b37d5.genome'):

    if not check_vcf_merge_compatibility(mity_vcf, hc_vcf):
        logging.error("The VCF files use mitochondrial contigs")
        sys.exit()

    prefix = create_prefix(hc_vcf, prefix)
    outfile = prefix + ".mity.vcf.gz"

    logging.debug("importing hc_vcf: " + hc_vcf)
    hc_file = gzip.open(hc_vcf, 'rt')
    
    # split the header and the variants into two separate lists
    hc_header = []
    hc_variants = []
    for line in hc_file:
        if line[0] == "#":
            line = line.strip()
            hc_header.append(line)
        else:
            line = line.strip()
            # line = line.split('\t')
            hc_variants.append(line)
    
    hc_col_names = hc_header[-1]
    del hc_header[-1]

    logging.debug("importing mity_vcf: " + mity_vcf)
    mity_file = gzip.open(mity_vcf, 'rt')
    
    # split the header and the variants into two separate lists
    # TODO: to speed up, do we actually need to get hc variants?
    mity_header = []
    mity_variants = []
    for line in mity_file:
        if line[0] == "#":
            line = line.strip()
            mity_header.append(line)
        else:
            line = line.strip()
            line = line.split('\t')
            mity_variants.append(line)
    
    mity_col_names = mity_header[-1]
    del mity_header[-1]
    
    # print(mity_header)
    # print(hc_header)
    
    # remove file format (e.g. 4.1/4.2) for mity and hc add manually to the 
    # merged header
    del mity_header[0]
    del hc_header[0]
    
    # remove phasing from mity header
    mity_header = [x for x in mity_header if "phasing=" not in x]
    
    # remove file date from mity header
    mity_header = [x for x in mity_header if "##fileDate" not in x]
    
    # get mity reference
    mity_ref = [x for x in mity_header if "##reference=" in x]
    
    # should only be one line in the mity header that has reference in it
    if len(mity_ref) > 1:
        err_msg = "Mity header has more than lines that define a reference: \n"
        for x in mity_ref:
            err_msg = err_msg + x
            err_msg = err_msg + '\n'
        sys.exit(err_msg)
    
    mity_ref = mity_ref[0].split('##reference=')[1]
    # print(mity_ref)
    # remove mity reference from header because we will add it with hc 
    # reference later
    mity_header = [x for x in mity_header if "##reference" not in x]
    
    # get hc reference then remove from header
    hc_ref = [x for x in hc_header if "##reference" in x]
    # should only be one line in the hc header that has reference in it
    if len(hc_ref) > 1:
        err_msg = "HC header has more than lines that define a reference: \n"
        for x in hc_ref:
            err_msg = err_msg + x
            err_msg = err_msg + '\n'
        sys.exit(err_msg)
    
    hc_ref = hc_ref[0].split('##reference=')[1]
    # remove hc reference from header because we will add it with mity 
    # reference later
    hc_header = [x for x in hc_header if "##reference" not in x]
    
    # make new reference line and add to merged header
    # @TODO add ##mity_reference, ##nuclear_reference, and set ##reference=hc_ref[0].split('##reference=')
    new_ref_line = "##reference=If CHR=MT: " + mity_ref + ". Otherwise: " + hc_ref
    
    # remove all the lines in the mity header that are also in the HC header
    # this should remove contig lines if they are the same
    mity_header = [x for x in mity_header if x not in hc_header]
    
    # now if there are two lines in mity_header and hc_header that 
    # that have the same eg "##INFO=<ID=SRR,", then they have different 
    # definitions for the same 
    # ID. need to make hc and mity specific definitions.
    sep = ","
    merged_header = []
    mity_ids = [x.split(sep)[0] for x in mity_header]
    # print(mity_ids)
    # loop through the hc_header
    for hc_line in hc_header:
        hc_id = hc_line.split(sep, 1)[0]
        if hc_id in mity_ids:
            # print(hc_line)
            mity_line_idx = mity_ids.index(hc_id)
            mity_line = mity_header[mity_line_idx]
            # print(mity_line)
            
            # check the number is the same - if not put "."
            mity_number = mity_line.split("Number=")[1]
            mity_number = mity_number.split(",")[0]
            # print(mity_number)
            
            hc_number = hc_line.split("Number=")[1]
            hc_number = hc_number.split(",")[0]
            # print(hc_number)
            
            if str(hc_number) == str(mity_number):
                new_number = hc_number
            else:
                new_number = "."
            
            # check the type is the same
            # if not will have to stop the app with an error
            # no way of saying unknown type
            # and they really should be the same anyway
            mity_type = mity_line.split("Type=")[1]
            mity_type = mity_type.split(",")[0]
            # print(mity_type)
            
            hc_type = hc_line.split("Type=")[1]
            hc_type = hc_type.split(",")[0]
            # print(hc_type)
            
            if str(hc_type) == str(mity_type):
                new_type = hc_type
            else:
                new_type = hc_type
            # print("")
            # sys.exit("TODO")
            
            hc_description = hc_line.split('Description="')[1]
            # remove the last ">
            hc_description = hc_description[:-2]
            # print(hc_description)
            
            mity_description = mity_line.split('Description="')[1]
            # remove the last ">
            mity_description = mity_description[:-2]
            # print(mity_description)
            
            new_header_line = hc_id + ",Number=" + new_number + ",Type=" + \
                              new_type + ",Description=\"If CHR=MT: \'" + \
                              mity_description + "\'. Otherwise: \'" + \
                              hc_description + "\' \">"
            # print(new_header_line)
            merged_header.append(new_header_line)
            
            # remove the line from the mity_header
            del mity_header[mity_line_idx]
        # print(len(mity_header))
        else:
            merged_header.append(hc_line)
    
    # now add the rest of the mity_header to the merge header
    merged_header = merged_header + mity_header
    merged_header = sorted(merged_header)
    merged_header.insert(0, new_ref_line)
    merged_header.insert(0, '##fileformat=VCFv4.1')
    
    # for line in merged_header:
    # 	print(line)
    
    # now check that the sample names are in the right order
    
    # get hc and mity sample names
    # assumes that sample names always start from 10th column
    hc_col_names = hc_col_names.split('\t')
    hc_samples = hc_col_names[9:]
    # print(hc_samples)
    
    mity_col_names = mity_col_names.split('\t')
    mity_samples = mity_col_names[9:]
    # print(mity_samples)
    
    if hc_samples == mity_samples:
        sys.stderr.write(
            "samples in same order - can just add mity variants to hc vcf")
        # add the hc variants with the mity variants - will be ordered later
        hc_col_names = ('\t').join(hc_col_names)
        
        new_mity_variants = []
        for line in mity_variants:
            new_line = ('\t').join(line)
            new_mity_variants.append(new_line)
        
        new_vcf = merged_header + [
            hc_col_names] + new_mity_variants + hc_variants
        for line in new_vcf:
            print(line)
    else:
        # check that they have the same samples
        if set(hc_samples) != set(mity_samples):
            sys.exit("VCFs have different sample names")
        mity_variants_no_sample = [x[:9] for x in mity_variants]
        mity_variants_sample = [x[9:] for x in mity_variants]
        # print(mity_variants_sample)
        # sys.stderr.write("samples not in the same order - will reorder mity
        # variants")
        
        # get the order of the hc sampels
        # print("hc samples")
        # print(hc_samples)
        # print("mity samples")
        # print(mity_samples)
        # mity_samples_idx = [hc_samples.index(x) for x in mity_samples]
        mity_samples_idx = [mity_samples.index(x) for x in hc_samples]
        
        # print(mity_samples_idx)
        # sys.exit()
        mity_variants_reordered_sample = []
        for line in mity_variants_sample:
            mity_variants_reordered_sample.append(
                    [line[x] for x in mity_samples_idx])
        # print(mity_variants_reordered_sample)
        # sys.exit()
        # add the newly ordered samples back onto the mity_variants_no_sample
        new_mity_variants = []
        for i in range(0, len(mity_variants_no_sample)):
            new_line = mity_variants_no_sample[i] + \
                       mity_variants_reordered_sample[i]
            # print(new_line)
            new_line = ('\t').join(new_line)
            new_mity_variants.append(new_line)
        
        # add the hc variants with the mity variants - will be ordered later
        hc_col_names = ('\t').join(hc_col_names)
        new_vcf = merged_header + [
            hc_col_names] + new_mity_variants + hc_variants

        write_merged_vcf(new_vcf, outfile, genome)



"""Normalise mity VCF."""

import sys
import gzip
import re
import logging
from .util import write_vcf
from scipy.stats import binom
from math import isinf
import numpy as np

def unchanged(List):
    # check that all numbers in the list are the same
    if len(set(List)) != 1:
        sys.exit("Multiple values in " + str(List))
    return ((List[0]))


def Sum(List):
    # convert to integer
    List = [float(i) for i in List]
    return (sum(List))


def weighted_mean(mean_list, count_list):
    mean_list = [float(i) for i in mean_list]
    count_list = [float(i) for i in count_list]

    numerator = [a * b for a, b in zip(mean_list, count_list)]
    numerator = sum(numerator)
    denominator = sum(count_list)
    return (numerator / denominator)


# this will split multiallelic lines in a vcf so that each allele is on a separate line
# variants is a list, each element is a line from the vcf
# eg variants = ['MT\t3103\t.\tCTACN\tCTAC,CTACT\t68191.9\t.\tAB=0,0;ABP=0,0;AC=2,4;AF=0.333333,0.666667;AN=6;;...,'MT\t3108\t.\tTT\tTC\t0\t.\tAB=0;ABP=0;AC=0;AF=0;AN=6;...]
def split_multi_allelic(variants):
    split_vcf = []

    # loop over each line in variants
    for line in variants:

        lineparts = line.split('\t')
        # print(lineparts)
        # lineparts = ['MT', '3103', '.', 'CTACN', 'CTAC,CTACT', '68191.9', '.', 'AB=0,0;ABP=0,0;AC=2,4;AF=0.333333,0.666667;AN=6;...]

        # The position of these fields should never change in a vcf:
        chromosome = lineparts[0]
        pos = lineparts[1]
        ID = lineparts[2]
        ref = lineparts[3]
        QUAL = lineparts[5]
        Filter = lineparts[6]

        alt = lineparts[4]
        # print(alt):
        # alt = CTAC,CTACT
        alt = alt.split(",")
        # print(alt):
        # alt = ['CTAC', 'CTACT']
        n_alleles = len(alt)

        ######################################
        ###### INFO field

        INFO = lineparts[7]
        INFO = INFO.split(";")
        # print(INFO):
        # INFO = ['AB=0,0', 'ABP=0,0', 'AC=2,4', 'AF=0.333333,0.666667', 'AN=6', ...]
        # get the abbreviated names in the info field
        INFO_names = [x.split("=")[0] for x in INFO]
        # print(INFO_names):
        # INFO_names = ['AB', 'ABP', 'AC', 'AF', 'AN'...]

        # INFO_to_keep are the INFO fields that you want to remain in the VCF.
        # Note that AF is replaced by VAF later, so don't need to keep this
        INFO_to_keep = ["DP", "QR", "QA", "SRF", "SRR", "SAF", "SAR", "MQM", "MQMR"]

        # get the index of where each of the INFO_to_keep are in the INFO
        # better way to do this?
        idx = []
        for keep_name in INFO_to_keep:
            counter = 0
            for names in INFO_names:
                # print(keep_name)
                if names == keep_name:
                    # print(names)
                    idx.append(counter)
                counter += 1
        idx = sorted(idx)
        # print(idx):
        # idx = [7, 15, 16, 26, 27, 34, 36, 37, 39]

        # pick out the info names, this should be the same as INFO_to_keep
        INFO_names = [INFO_names[i] for i in idx]
        # add the equal sign back in
        INFO_names = [x + "=" for x in INFO_names]
        # print(INFO_names):
        # INFO_names = ['DP=', 'MQM=', 'MQMR=', 'QA=', 'QR=', 'SAF=', 'SAR=', 'SRF=', 'SRR=']

        # get the values of the INFO
        INFO_values = [x.split("=")[1] for x in INFO]
        INFO_values = [INFO_values[i] for i in idx]
        # print(INFO_values)
        # INFO_values = ['7575', '70,60', '0', '75397,5623', '0', '1289,100', '1351,103', '0', '0']

        # split the values for each allele
        INFO_values = [x.split(",") for x in INFO_values]
        # print(INFO_values)
        # INFO_values = [['7575'], ['70', '60'], ['0'], ['75397', '5623'], ['0'], ['1289', '100'], ['1351', '103'], ['0'], ['0']]

        # get the number of values for each allele
        len_INFO_values = [len(x) for x in INFO_values]
        # print(len_INFO_values)
        # len_INFO_values = [1, 2, 1, 2, 1, 2, 2, 1, 1]

        # In INFO_values, there should only be two lengths of the sublists,
        # either length 1, if its a reference value, or length of the number of alleles
        if set(len_INFO_values) != set([1, n_alleles]):
            sys.exit(
                "the values in the info field must either be for the reference (length = one) or for the alternates (length = number of alleles)")

        # if there is only one value, i.e. just for the reference, repeat it the number of alleles times
        # so that it can be picked out later when each allele is put on a new line
        INFO_values = [x * n_alleles if len(x) == 1 else x for x in INFO_values]
        # info_values = [x*n_alleles if len(x)<n_alleles else x for x in info_values]
        # print(INFO_values)
        # INFO_values = [['7575', '7575'], ['70', '60'], ['0', '0'], ['75397', '5623'], ['0', '0'], ['1289', '100'], ['1351', '103'], ['0', '0'], ['0', '0']]

        ######################################
        ###### GENO/FORMAT field
        FORMAT_name = lineparts[8]
        FORMAT_name = FORMAT_name.split(":")
        logging.debug(FORMAT_name)
        # FORMAT_name = ['GT', 'DP', 'AD', 'RO', 'QR', 'AO', 'QA', 'GL']

        FORMAT_values = lineparts[9:]
        # print(FORMAT_values)
        #  = ['2/2:2915:0,0,88:0:0:0,88:0,2441:-251.884,-310.442,-249.634,-90.9111,-88.9622,0', '2/2:1936:0,0,67:0:0:0,67:0,1833:-174.582,-231.477,-174.582,-78.2678,-78.2678,0', '1/1:2724:0,2640,48:0:0:2640,48:75397,1349:-6804.65,-872.084,0,-6770.44,-762.912,-6680.73']

        FORMAT_values = [x.split(":") for x in FORMAT_values]
        # print(FORMAT_values)

        # remove GL as we dont really know what to do with it yet
        try:
            idx = FORMAT_name.index("GL")
            del FORMAT_name[idx]
            for x in FORMAT_values:
                del x[idx]
            # print(FORMAT_values)
            # FORMAT_values = [['2/2', '2915', '0,0,88', '0', '0', '0,88', '0,2441'], ['2/2', '1936', '0,0,67', '0', '0', '0,67', '0,1833'], ['1/1', '2724', '0,2640,48', '0', '0', '2640,48', '75397,1349']]
        except ValueError:
            pass
        n_samples = len(FORMAT_values)
        # print(n_samples)

        ##################################
        ####### Create new line

        for allele in range(0, n_alleles):

            # build up the new line
            new_line = [chromosome, pos, ID, ref, alt[allele], QUAL, Filter]

            # build up new INFO field
            # will not be affected by number of samples
            # pick out the allele'th value from each of the info values
            new_info_values = [x[allele] for x in INFO_values]
            # create empty list that will be the new info field
            new_info = [None] * (len(new_info_values) + len(INFO_names))
            # fill it with the info names and new values
            new_info[::2] = INFO_names
            new_info[1::2] = new_info_values
            # paste the name with the corresponding value (which will be directly
            # after the name in the list)
            si = iter(new_info)
            new_info = [x + next(si, '') for x in si]
            # print(new_info)
            # join all of the info fields with ;
            new_info = ";".join(new_info)

            # add new INFO field to new_line
            new_line.append(new_info)

            # add the FORMAT names to the new line
            new_FORMAT_name = ":".join(FORMAT_name)
            new_line.append(new_FORMAT_name)
            # print(new_line)
            # new_line = ['MT', '3103', '.', 'CTACN', 'CTAC', '68191.9', '.', 'DP=7575;MQM=70;MQMR=0;QA=75397;QR=0;SAF=1289;SAR=1351;SRF=0;SRR=0', 'GT:DP:AD:RO:QR:AO:QA']
            # new_line = ['MT', '3103', '.', 'CTACN', 'CTACT', '68191.9', '.', 'DP=7575;MQM=60;MQMR=0;QA=5623;QR=0;SAF=100;SAR=103;SRF=0;SRR=0', 'GT:DP:AD:RO:QR:AO:QA']

            # Add new FORMAT field
            # FORMAT field is affected by number of samples

            for samp in range(0, n_samples):
                temp_format = FORMAT_values[samp]
                logging.debug(temp_format)

                # split the allele values in the FORMAT
                temp_format = [x.split(",") for x in temp_format]
                # if temp_format[0] == ['2/2']:
                # print(temp_format)
                # print(temp_format)

                # GT will be overwritten later
                # DP will only have one value per sample, and will be unaffected by allele
                # RO will only have one value per sample, and will be unaffected by allele
                # QR will only have one value per sample, and will be unaffected by allele
                #
                # AD, AO and QA need to be updated - depend on allele

                # AO_idx is where AO sits in the FORMAT field
                AO_idx = FORMAT_name.index("AO")
                # pick out the allele'th value of AO
                new_AO = temp_format[AO_idx][allele]
                # print(new_AO)

                # update the format field
                # print(temp_format)
                temp_format[AO_idx] = [new_AO]
                # print(temp_format)
                # print("")

                # QA_idx is where QA sits in the FORMAT field
                QA_idx = FORMAT_name.index("QA")
                # pick out the allele'th value of QA
                new_QA = temp_format[QA_idx][allele]
                # update the format field
                temp_format[QA_idx] = [new_QA]

                # AD_idx is where AD sits in the FORMAT field
                AD_idx = FORMAT_name.index("AD")
                # pick out the allele'th + 1 value of AD
                # +1 because the first value is always the reference depth
                new_AD = temp_format[AD_idx][allele + 1]
                # add the reference depth in
                new_AD = [temp_format[AD_idx][0], new_AD]
                # print(new_AD)
                # update the format field
                temp_format[AD_idx] = new_AD
                # print(temp_format)
                temp_format = [",".join(x) for x in temp_format]
                # print(temp_format)
                temp_format = ":".join(temp_format)
                # print(temp_format)

                # add the new format field to the end of the new line
                new_line.append(temp_format)
                # print(new_line)

            split_vcf.append(new_line)

    return (split_vcf)


# variant_list is a list, with no multiallelic variants
# Eg variant_list = [['MT', '3103', '.', 'CTACN', 'CTAC', '68191.9', '.', 'DP=7575; ... ], ['MT', '3103', '.', 'CTACN', 'CTACT', '68191.9', '.', 'DP=7575;]]
def split_MNP(variant_list):
    table = []

    for line in variant_list:
        # print(line)
        ref = list(line[3])
        ref_length = len(ref)

        alt = list(line[4])
        alt_length = len(alt)

        if alt_length > 1 and ref_length > 1:
            # then this is a MNP

            if alt_length != ref_length:
                # then ref and alt dont have the ame number of bases
                # there isnt much we can do to simplify apart from trim the starting bases if they are matching
                # could be a complex variant if it doesnt simplify after trimming

                min_length = min(alt_length, ref_length)
                num_matching_bases = 0
                # only loop over the minimum of alt or ref bases
                # count how many of the bases at the start are the same
                for base in range(0, min_length):
                    if ref[base] != alt[base]:
                        break
                    else:
                        num_matching_bases += 1
                # trim_num is the number of bases at the start to trim
                trim_num = num_matching_bases - 1
                temp_line = list(line)
                if trim_num > 0:
                    # update the position
                    temp_line[1] = int(line[1]) + trim_num
                    # trim the reference
                    temp_line[3] = temp_line[3][trim_num:]
                    # trim the alternate
                    temp_line[4] = temp_line[4][trim_num:]

                # if the new reference and alternate length are still > 1 then
                # then this is defined as a complex variant
                if len(temp_line[3]) > 1 and len(temp_line[4]) > 1:
                    # print("its still a complex variant")
                    temp_line[7] = temp_line[7] + ";TYPE=complex"
                    # print(temp_line)

                table.append(temp_line)

            else:
                # the alternate and the reference have the same number of bases
                # so we just try to remove the extra bases
                # This logic could be changed so that variants next to each other are reported together
                # eg AT -> CG is currently reported A->C and T->G
                for base_idx in range(0, ref_length):

                    if alt[base_idx] != ref[base_idx]:
                        # then this is a variant
                        temp_line = list(line)
                        # update the position, reference and alternate
                        temp_line[1] = int(line[1]) + base_idx
                        temp_line[3] = ref[base_idx]
                        temp_line[4] = alt[base_idx]

                        table.append(temp_line)
        else:
            # If either the alt or reference is one base, then not an MNP
            table.append(line)

    return (table)


# variant_list is a list of sublists, each sublist is  line in the vcf
# variant_list is a list, with no multiallelic variants
# Eg variant_list = [['MT', 73.0, '.', 'A', 'G', '148165', '.', 'DP=5912...], ['MT', 146.0, '.', 'T', 'C', '198193', '.', 'DP=7697;...], ['MT', '182', '.', 'C', 'T', '6.00899e-14', '.', 'DP=7955;...]]
def combine_lines(variant_list, p=0.002):
    # make dictionary with keys chromosome, position and alternate

    # print(variant_list)
    # we are interested in the lines that have the same chromosome position and alternate -
    # to identify these lines, create an "id" for each line that is position_alternate
    # if any lines have matching ids then they should be combined.

    # get the chromosome positions and alternates

    chromo = [x[0] for x in variant_list]
    unique_chromosomes_temp = list(set(chromo))

    # trying to keep the variants in a rough order (set doesnt preserve order)
    autosome = [str(x) for x in list(range(1, 23))]
    unique_chromosomes = [x for x in autosome if x in unique_chromosomes_temp]
    unique_chromosomes = unique_chromosomes + [x for x in ['X', 'Y', 'MT'] if x in unique_chromosomes_temp]
    unique_chromosomes = unique_chromosomes + [x for x in unique_chromosomes_temp if x not in unique_chromosomes]

    # print(unique_chromosomes)
    # sys.exit()
    # unique_chromosomes = [str(x) for x in unique_chromosomes]
    # x = list(range(1,23))
    # print(x)
    new_vcf = []
    # we loop over the chromsomes to fill new_vcf
    for uniq_chrom in unique_chromosomes:
        logging.debug('Processing chromosome ' + str(uniq_chrom))
        chromo_variant_list = [line for line in variant_list if line[0] == uniq_chrom]
        # print(chromo_variant_list)
        # sys.exit()
        # TODO: could remove the lines from the variant list now - would make faster?
        # print(uniq_chrom)
        # print(chromo_variant_list)
        # sys.exit()
        position = [x[1] for x in chromo_variant_list]
        # print(position)
        alternate = [x[4] for x in chromo_variant_list]

        # paste the position and alternate to make id called pos_alt_id
        line_id = []
        for pos, alt in zip(position, alternate):
            line_id.append("_".join([str(pos), alt]))
        # print(line_id)
        # line_id = ['MT_73_G', 'MT_116_C', 'MT_139_G', 'MT_146_C', ... , 'MT_16263_C', 'MT_16264_T', 'MT_16263_C', 'MT_16270_T', ...]

        # line_number is the line number from chromo_variant_list
        # this is to keep track of whats already been added to the new vcf
        line_number = list(range(len(chromo_variant_list)))
        # print(line_number)
        # line_number = [0, 1, 2, 3, ... 497, 498, 499, 500]
        while len(line_id) > 0:
            # take the first id
            x = line_id[0]
            # print(x)
            # x = 73_G
            # or later in the loop
            # x = 16263_C

            # find all the matching ids, including the first
            # print("finding matching lines")
            matching_lines_idx = [i for i, j in enumerate(line_id) if j == x]
            # print("finished findinf matching lines")
            # print(matching_lines_idx)
            # matching_lines_idx = [0] i.e. there is only one 73_G
            # or furher into the loop
            # matching_lines_idx = [0,2] i.e. there are 2 16263_C

            matching_lines = [line_number[i] for i in matching_lines_idx]
            # print(matching_lines)
            # matching_lines = [0]
            # or later into the loop for 16263_C
            # matching_lines = [497, 499]

            # remove the line_id and the line_number that have been assigned to matching_lines
            for i in sorted(matching_lines_idx, reverse=True):
                del line_id[i]
            for i in sorted(matching_lines_idx, reverse=True):
                del line_number[i]
                # print(line_id)
            # line_id = ['116_C', '139_G', '146_C', '152_C', '182_T', .... ]
            # or later into the loop
            # if x == '16263_C':
            # print(line_id)
            # line_id = ['16264_T', '16270_T', '16274_A', '16278_T', '16284_G', '16288_C', '16290_T', '16293_C', '16301_T', '16311_C', '16343_C', ... ]
            # ie 16263_C has been removed everywhere in list

            # print(line_number)
            # line_number = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, ... ]
            # or later into the loop
            # if x == '16263_C':
            # print(line_number)
            # line_number = [498, 500, 501, 502, 503 ... ]

            # TODO: the way I manipulate the INFO line is different in each case
            # should probably make the same so its easier to follow
            if len(matching_lines) == 1:
                # print("not repeated position")
                # print(matching_lines)
                # This is a unique record for this genomic position.
                # we just need to update and add values to the info/format field.

                # pick out the line from chromo_variant_list:
                # need the [0] to remove it from list, better way to do this? ie [[x]] to [x]
                new_line = [chromo_variant_list[i] for i in matching_lines][0]
                # print(new_line)

                INFO = new_line[7]
                INFO = INFO.split(";")

                INFO_names = [i.split("=")[0] for i in INFO]
                # print(INFO_names)
                INFO_values = [i.split("=")[1] for i in INFO]

                # we want to caclulate SBR and add to INFO
                # SBR = strand bias of reference
                # SBR = SRF/(SRF+SRR)

                SRF_idx = INFO_names.index('SRF')
                SRF = float(INFO_values[SRF_idx])

                SRR_idx = INFO_names.index('SRR')
                SRR = float(INFO_values[SRR_idx])

                # to avoid dividing by zero:
                if SRF > 0 or SRR > 0:
                    SBR = round(SRF / (SRF + SRR), 3)
                else:
                    SBR = 0

                INFO_values.append(SBR)
                INFO_names.append("SBR")

                # we want to calculate SBA and add to INFO
                # SBA = strand bias of alternate
                # i.e. SBA = SAF/(SAF+SAR)

                SAF_idx = INFO_names.index('SAF')
                SAF = float(INFO_values[SAF_idx])

                SAR_idx = INFO_names.index('SAR')
                SAR = float(INFO_values[SAR_idx])
                # note SAF will always > 0, otherwise variant wouldn't be reported
                SBA = round(SAF / (SAF + SAR), 3)

                INFO_values.append(SBA)
                INFO_names.append("SBA")

                # we want to add VAF to the FILTER field
                # to do this we need AO and DP
                # we also want to update the genotype
                # and add AQR - average base quality of the reference
                # we update AD to be RO,AO, because of the bug in earlier version that had AD wrong, just to be safe
                # these are all per sample

                # because not matching position, there will only be one line of FORMAT values
                FORMAT_values = new_line[9:]
                FORMAT_values = [x.split(":") for x in FORMAT_values]
                # print(FORMAT_values)
                n_samples = len(FORMAT_values)

                FORMAT_names = new_line[8]
                FORMAT_names = FORMAT_names.split(":")

                # add AQR after QR
                QR_idx = FORMAT_names.index('QR')
                FORMAT_names.insert((QR_idx + 1), 'AQR')
                # add AQA after QA
                QA_idx = FORMAT_names.index('QA')
                FORMAT_names.insert((QA_idx + 1), 'AQA')
                # print(FORMAT_names)

                # add VAF to the format
                FORMAT_names.append('VAF')
                # add q to the formal
                FORMAT_names.append('q')
                # print(FORMAT_names)
                # add dummy VAF placeholder
                [x.extend(['-1']) for x in FORMAT_values]
                # add dummy q placeholder
                [x.extend(['-1']) for x in FORMAT_values]
                # print(FORMAT_values)

                AO_vector = []
                # VAF = []
                DP_vector = []
                # new_genotype = []
                for samp in range(0, n_samples):
                    # print("")
                    temp_format = FORMAT_values[samp]
                    # print(temp_format)
                    # add dummy placeholder for AQR and AQA
                    temp_format.insert((QR_idx + 1), '-1')
                    temp_format.insert((QA_idx + 1), '-1')
                    # print(temp_format)

                    AO_idx = FORMAT_names.index('AO')
                    AO = int(temp_format[AO_idx])
                    AO_vector.append(AO)
                    # print(AO)
                    # print(AO_vector)

                    RO_idx = FORMAT_names.index('RO')
                    RO = int(temp_format[RO_idx])

                    # change the format DP to be AO + RO
                    DP_idx = FORMAT_names.index('DP')
                    DP = RO + AO
                    temp_format[DP_idx] = str(DP)
                    DP_vector.append(DP)
                    # DP = int(temp_format[DP_idx])
                    # print(DP)

                    # DP = AO + RO can be zero if another sample cal
                    if DP != 0:
                        samp_VAF = round(float(AO) / float(DP), 4)
                    else:
                        samp_VAF = 0

                    # VAF.append(str(samp_VAF))
                    VAF_idx = FORMAT_names.index('VAF')
                    temp_format[VAF_idx] = str(samp_VAF)

                    # add in q
                    q = mity_qual(AO, DP, p=p)
                    q_idx = FORMAT_names.index('q')
                    temp_format[q_idx] = str(q)

                    # print(p)
                    # print(float(DP))
                    # print(float(AO))
                    # print(q)
                    # exit()
                    # add in the new genotype
                    if AO < 4:
                        new_genotype = "0/0"
                    else:
                        if samp_VAF == 1:
                            new_genotype = "1/1"
                        elif samp_VAF < 1:
                            new_genotype = "0/1"

                    # print(VAF)
                    # if samp_VAF > 0.9:
                    #   new_genotype = "1/1"
                    # elif samp_VAF > 0:
                    #   new_genotype = "0/1"
                    # else:
                    #   new_genotype = "0/0"
                    # print(new_genotype)

                    # replace the genotypes
                    GT_idx = FORMAT_names.index('GT')
                    temp_format[GT_idx] = new_genotype
                    # print(temp_format)
                    # print("")

                    RO_idx = FORMAT_names.index('RO')
                    RO = int(temp_format[RO_idx])
                    # print(RO)

                    # calculate new AD
                    new_AD = [str(RO), str(AO)]
                    new_AD = ",".join(new_AD)
                    AD_idx = FORMAT_names.index('AD')
                    temp_format[AD_idx] = new_AD
                    # print(temp_format)

                    # calculate AQR
                    QR_idx = FORMAT_names.index('QR')
                    # print(AO_idx)
                    QR = float(temp_format[QR_idx])
                    if RO > 0:
                        AQR = str(round(QR / RO, 3))
                    else:
                        # if there are no reference reads, set AQR high so you dont filter on it
                        # AQR = 1000000
                        # AQR = 1000000 will be confusing, so check in the filters function if RO = 0
                        AQR = str(0)
                    # add AQR to the FORMAT field
                    AQR_idx = FORMAT_names.index('AQR')
                    temp_format[AQR_idx] = AQR

                    # calculate AQA
                    QA_idx = FORMAT_names.index('QA')
                    QA_idx = FORMAT_names.index('QA')
                    # print(AO_idx)
                    QA = float(temp_format[QA_idx])
                    # here AO could be less than zero because it could be another sample that has this variant
                    if AO > 0:
                        AQA = str(round(QA / AO, 3))
                    else:
                        AQA = str(0)
                    # add AQA to the FORMAT field
                    AQA_idx = FORMAT_names.index('AQA')
                    temp_format[AQA_idx] = str(AQA)

                    # add the updated FORMAT field for the given sample to the new_line
                    temp_format = ":".join(temp_format)
                    new_line[samp + 9] = temp_format

                # update QUAL
                cohort_AO = sum(int(x) for x in AO_vector)
                cohort_DP = sum(int(x) for x in DP_vector)
                cohort_QUAL = mity_qual(cohort_AO, cohort_DP, p=p)
                new_line[5] = str(round(cohort_QUAL, 1))
                logging.debug("Cohort QUAL score: {} := {} vs {}".format(new_line[5], ",".join(str(x) for x in AO_vector), ",".join(str(x) for x in DP_vector)))

                # add the new FORMAT names
                FORMAT_names = ":".join(FORMAT_names)
                new_line[8] = FORMAT_names
                # print(new_line)

                # add VAF to the info
                # VAF = ",".join(VAF)
                # INFO_values.append(VAF)
                # INFO_names.append("VAF")

                # dont update the INFO depth - doesnt make sense
                # when you have more than one sample
                # unless you do RO+AO for each sample
                # ie total depth at that position in all the samples...

                INFO_DP_idx = INFO_names.index('DP')
                new_info_DP = sum(DP_vector)
                INFO_values[INFO_DP_idx] = new_info_DP
                # print(DP_vector)
                # print(new_info_DP)

                # put the = sign back
                INFO_names = [x + "=" for x in INFO_names]
                new_info = [None] * (len(INFO_names) + len(INFO_values))
                new_info[::2] = INFO_names
                new_info[1::2] = INFO_values
                new_info = [str(x) for x in new_info]
                si = iter(new_info)
                new_info = [x + next(si, '') for x in si]
                new_info = ";".join(new_info)
                # print(new_info)
                # update INFO
                new_line[7] = new_info
                # print(new_line)
                new_vcf.append(new_line)


            else:
                # then this is a repeated position
                # we need to update and add values to the info/format field, as well as combine the matching lines
                # print("repeated position")
                # print(matching_lines)

                new_line = [chromo_variant_list[i] for i in matching_lines]
                replacement_line = new_line[0]
                # print(new_line)

                #######################
                ##### FORMAT field
                new_FORMAT = []

                # loop over samples, and update each sample separately and then add to new format
                # each line should have same number of samples
                # TODO could move the calculation of the number of samples to the start of the function
                first_FORMAT_values = new_line[0][9:]
                n_samples = len(first_FORMAT_values)

                # FORMAT_values = new_line[9:]
                # FORMAT_values = [x.split(":") for x in FORMAT_values]
                # print(FORMAT_values)

                FORMAT_names = new_line[0][8]
                FORMAT_names = FORMAT_names.split(":")
                # print(FORMAT_names)

                # we need the AO vector later for weighted mean
                AO_idx = FORMAT_names.index('AO')
                AO_sum_vector = []
                for line in new_line:
                    # print(line)
                    formats = line[9:]
                    formats = [x.split(":") for x in formats]
                    AOs = [float(x[AO_idx]) for x in formats]
                    AO_sum_vector.append(sum(AOs))
                # print(AO_sum_vector)

                # add AQR after QR
                QR_idx = FORMAT_names.index('QR')
                FORMAT_names.insert((QR_idx + 1), 'AQR')
                # add AQA after QA
                QA_idx = FORMAT_names.index('QA')
                FORMAT_names.insert((QA_idx + 1), 'AQA')
                # print(FORMAT_names)
                # add the updated FORMAT_names to the new_line

                # add VAF to the format
                FORMAT_names.append('VAF')
                # add q to the format
                FORMAT_names.append('q')
                # print(FORMAT_names)
                # add dummy VAF placeholder
                # [x.extend(['-1']) for x in FORMAT_values]
                # print(FORMAT_values)

                new_AD = []
                VAF = []
                # because we need it later, create and keep an new_AO vector
                new_AO_vector = []
                new_DP_vector = []
                for samp in range(0, n_samples):
                    #   # print("")
                    temp_format = [i[9 + samp].split(":") for i in new_line]
                    # print(temp_format)

                    # add dummy placeholder for AQR and AQA

                    for x in range(0, len(temp_format)):
                        temp_format[x].insert((QR_idx + 1), '-1')
                        temp_format[x].insert((QA_idx + 1), '-1')
                    # print(temp_format)

                    # AO
                    AO_idx = FORMAT_names.index('AO')
                    # TODO: do we need new_AO_vector anymore?
                    AO_vector = [i[AO_idx] for i in temp_format]
                    new_AO = str(int(Sum(AO_vector)))
                    new_AO_vector.append(new_AO)
                    # new_AO = str(int(new_AO))
                    # print(AO_vector)
                    # print(new_AO)

                    # RO
                    RO_idx = FORMAT_names.index('RO')
                    RO_vector = [i[RO_idx] for i in temp_format]
                    new_RO = unchanged(RO_vector)
                    new_RO = str(int(new_RO))
                    # print(RO_vector)
                    # print(new_RO)

                    new_AD = [str(new_RO), str(new_AO)]
                    new_AD = ",".join(new_AD)

                    # DP
                    DP_idx = FORMAT_names.index('DP')
                    new_DP = str(int(new_AO) + int(new_RO))
                    new_DP_vector.append(new_DP)
                    # DP_vector = [i[DP_idx] for i in temp_format]
                    # new_DP = unchanged(DP_vector)
                    # new_DP = str(int(new_DP))
                    # print(DP_vector)
                    # print(new_DP)

                    if new_DP != 0:
                        samp_VAF = round(float(new_AO) / float(new_DP), 4)
                    else:
                        samp_VAF = 0.0
                    logging.debug("Variant has samp_VAF: {}".format(str(samp_VAF)))
                    VAF.append(str(samp_VAF))
                    # print(VAF)

                    samp_q = mity_qual(new_AO, new_DP, p=p)
                    samp_q = str(samp_q)

                    # update GT
                    if int(new_AO) < 4:
                        new_genotype = "0/0"
                    else:
                        if samp_VAF >= 0.95:
                            new_genotype = "1/1"
                        elif samp_VAF < 0.95:
                            new_genotype = "0/1"
                    samp_VAF = str(samp_VAF)

                    # QR
                    QR_vector = [i[FORMAT_names.index('QR')] for i in temp_format]
                    new_QR = unchanged(QR_vector)
                    new_QR = str(new_QR)

                    # calculate AQR
                    if int(new_RO) > 0:
                        AQR = str(float(new_QR) / float(new_RO))
                    else:
                        # if there are no reference reads, set AQR high so you dont filter on it
                        # AQR = 1000000
                        # AQR = 1000000 will be confusing, so check in the filters function if RO = 0
                        AQR = str(0)

                    # QA
                    QA_idx = FORMAT_names.index('QA')
                    QA_vector = [i[QA_idx] for i in temp_format]
                    new_QA = Sum(QA_vector)
                    new_QA = str(int(new_QA))

                    # calculate AQA
                    # here AO could be less than zero because it could be another sample that has this variant
                    if int(new_AO) > 0:
                        AQA = str(float(new_QA) / float(new_AO))
                    else:
                        # if there are no reference reads, set AQR high so you dont filter on it
                        # AQR = 1000000
                        # AQR = 1000000 will be confusing, so check in the filters function if RO = 0
                        AQA = str(0)

                    new_FORMAT.append(":".join(
                        [new_genotype, new_DP, new_AD, new_RO, new_QR, AQR, new_AO, new_QA, AQR, samp_VAF, samp_q]))
                    # print(new_FORMAT)

                # print(new_FORMAT)

                #######################
                ##### INFO field
                INFO_field = [i[7].split(":") for i in new_line]
                # print(INFO_field)
                INFO_field = [x[0].split(";") for x in INFO_field]
                # print(INFO_field)

                INFO_values = []
                for line in INFO_field:
                    INFO_values.append([x.split("=")[1] for x in line])

                # print(INFO_values)

                INFO_names = INFO_field[0]
                INFO_names = [x.split("=")[0] for x in INFO_names]
                # print(INFO_names)

                # DP
                DP_idx = INFO_names.index('DP')
                # print(DP_idx)
                DP_vector = [i[DP_idx] for i in INFO_values]
                # print(DP_vector)
                new_DP = unchanged(DP_vector)
                # new_DP = sum(DP_vector)
                new_DP = "DP=" + str(new_DP)
                # print(new_DP)

                # MQM - Mean mapping quality of observed alternate alleles
                MQM_idx = INFO_names.index('MQM')
                # print(MQM_idx)
                MQM_vector = [i[MQM_idx] for i in INFO_values]
                new_MQM = weighted_mean(mean_list=MQM_vector, count_list=AO_sum_vector)
                new_MQM = "MQM=" + str(new_MQM)
                # print(new_MQM)

                # MQMR - Mean mapping quality of observed reference alleles
                MQMR_idx = INFO_names.index('MQMR')
                MQMR_vector = [i[MQMR_idx] for i in INFO_values]
                # print(MQMR_vector)
                new_MQMR = unchanged(MQMR_vector)
                new_MQMR = "MQMR=" + str(new_MQMR)
                # print(new_MQMR)

                # QA - Sum of quality of the alternate observations
                QA_idx = INFO_names.index('QA')
                QA_vector = [i[QA_idx] for i in INFO_values]
                new_QA = Sum(QA_vector)
                new_QA = "QA=" + str(new_QA)

                # QR - Sum of quality of the reference observations
                QR_idx = INFO_names.index('QR')
                QR_vector = [i[QR_idx] for i in INFO_values]
                new_QR = unchanged(QR_vector)
                new_QR = "QR=" + str(new_QR)

                # SAF - Number of alternate observations on the forward strand
                SAF_idx = INFO_names.index('SAF')
                SAF_vector = [i[SAF_idx] for i in INFO_values]
                new_SAF = Sum(SAF_vector)

                # SAR - Number of alternate observations on the reverse strand
                SAR_idx = INFO_names.index('SAR')
                SAR_vector = [i[SAR_idx] for i in INFO_values]
                new_SAR = Sum(SAR_vector)

                SBA = new_SAR / (new_SAR + new_SAF)

                # SRF - Number of reference observations on the forward strand
                SRF_idx = INFO_names.index('SRF')
                SRF_vector = [i[SRF_idx] for i in INFO_values]
                new_SRF = unchanged(SRF_vector)
                # print(new_SRF)

                # SRR - Number of reference observations on the forward strand
                SRR_idx = INFO_names.index('SRR')
                SRR_vector = [i[SRR_idx] for i in INFO_values]
                new_SRR = unchanged(SRR_vector)
                #  TODO: why do i need int here?
                if int(new_SRF) > 0 or int(new_SRR) > 0:
                    SBR = int(new_SRF) / (int(new_SRF) + int(new_SRR))
                else:
                    SBR = 0

                new_SAF = "SAF=" + str(new_SAF)
                new_SAR = "SAR=" + str(new_SAR)
                new_SRF = "SRF=" + str(new_SRF)
                new_SRR = "SRR=" + str(new_SRR)
                new_SBA = "SBA=" + str(SBA)
                new_SBR = "SBR=" + str(SBR)

                if "TYPE" in INFO_names:
                    # then add normalised to it
                    new_type = "TYPE=complex,normalised"
                else:
                    new_type = "TYPE=normalised"

                new_info = [new_DP, new_MQM, new_MQMR, new_QA, new_QR, new_SAF, new_SAR, new_SRF, new_SRR, new_type,
                            new_SBA, new_SBR]
                # add the VAF to the end of the info line
                # VAF = ",".join(VAF)
                # VAF = "VAF="+str(VAF)
                # new_info.append(VAF)
                new_info = ";".join(new_info)

                cohort_AO = sum(int(x) for x in new_AO_vector)
                cohort_DP = sum(int(x) for x in new_DP_vector)
                cohort_QUAL = mity_qual(cohort_AO, cohort_DP, p=p)
                logging.debug("Cohort QUAL score: {} := {} vs {}".format(replacement_line[5], ",".join(str(x) for x in new_AO_vector), ",".join(str(x) for x in new_DP_vector)))
                replacement_line[5] = str(round(cohort_QUAL, 1))
                replacement_line[7] = new_info
                replacement_line[9:] = new_FORMAT
                FORMAT_names = ":".join(FORMAT_names)
                replacement_line[8] = FORMAT_names
                # print(replacement_line)
                new_vcf.append(replacement_line)
    # print(new_vcf)
    return (new_vcf)


def mity_qual(AO, DP, p=0.002):
    """
    Compute variant quality
    :param AO: (int) number of alternative reads
    :param DP: (int) total read depth
    :param p: (float) noise parameter. default=0.002
    :return: (float) phred-scaled quality score

    >>> [mity_qual(x,10) for x in [1,2,3,4,5,6,7,8,9,10]]
    '[37.49, 60.22, 84.78, 110.97, 138.75, 168.16, 199.4, 232.92, 269.9, 296.89]'
    >>> [mity_qual(x,100) for x in [5,10,15,20,25,30,35,40,45,50]]
    '[71.87, 156.08, 251.23, 354.34, 463.9, 579.05, 699.21, 824.04, 953.32, 1086.94]'
    >>> [mity_qual(x,1000) for x in [5,10,15,20,25,30,35,40,45,50]]
    '[17.84, 50.97, 93.59, 142.89, 197.36, 256.02, 318.25, 383.57, 451.61, 522.1]'
    >>> [mity_qual(x,10000) for x in [5,10,15,20,25,30,35,40,45,50]]
    '[0.0, 0.05, 0.74, 3.56, 9.51, 18.73, 31.0, 46.04, 63.57, 83.37]'
    >>> mity_qual(118,118)
    '3211.76'
    >>> mity_qual(119,119)
    '3220'
    >>> mity_qual(10000,10000)
    '3220'
    """
    q = 0.0
    AO = int(AO)
    DP = int(DP)
    if AO > 0 and DP > 0:
        if DP == AO:
            # then penalise homoplasmic variants with low numbers of reads
            DP = DP + 1
        # v1: the cdf implementation caps at 159.55, due to the cdf capping at 0.9999999999999999
        # q = round(abs(-10 * log10(1 - binom.cdf(AO, DP, p))), 2)
        # v2: the logsf implementation is identical to cdf for low values and continues to scale up
        with np.errstate(divide='ignore'):
            q = float(round(-4.342945 * binom.logsf(AO, DP, p), 2))
    if isinf(q):
        # this is the maximum q that we observed looking at homoplastic variants around ~129x depth
        q = float(3220)
    return q

def add_filter(variant_list, min_DP=15, SB_range=[0.1, 0.9], min_MQMR=30, min_AQR=20):
    """
    FILTER out poor quality variants in a VCF.
    If one samples in the vcf passes, then they all pass in the FILTER column.

    :param variant_list: the parsed VCF
    :param min_DP: minimum total read depth of the variant
    :param SB_range: min and max SB range, where SB is the strand bias of either the REF (SBR_FAIL), or ALT (SBA_FAIL) reads.
    :param min_MQMR: minimum MQMR, where MQMR is the "Mean mapping quality of observed reference alleles"
    :param min_AQR: minimum MQMR, where MQMR is the "Mean mapping quality of observed reference alleles"
    :return: a
    """
    # @TODO: refactor this to use pyvcf

    # print(variant_list)
    # sys.exit()

    #############################################
    ##### HARD CODED FILTER VALUES
    blacklist = list(range(302, 319))
    blacklist = blacklist + [3105, 3106, 3107, 3108]

    new_vcf = []

    for line in variant_list:
        # print(line)

        POS = int(line[1])

        info = line[7].split(";")

        info_names = [x.split("=")[0] for x in info]
        info_values = [x.split("=")[1] for x in info]

        SBR = float(info_values[info_names.index("SBR")])
        SBA = float(info_values[info_names.index("SBA")])
        MQMR = float(info_values[info_names.index("MQMR")])

        FORMAT_names = line[8].split(":")

        FORMAT = line[9:]

        samp_POS_fil = []
        samp_SBR_fil = []
        samp_SBA_fil = []
        samp_MQMR_fil = []
        samp_AQR_fil = []
        # @clare, why don't you just fail variants with depth < min_DP in one filter (samp_DP_fil)?

        for samp in FORMAT:
            samp_FORMAT = samp.split(":")

            AO = int(samp_FORMAT[FORMAT_names.index("AO")])
            RO = int(samp_FORMAT[FORMAT_names.index("RO")])
            AQR = float(samp_FORMAT[FORMAT_names.index("AQR")])

            from .util import make_hgvs
            hgvs = make_hgvs(int(line[1]), line[3], line[4])
            if RO > min_DP:
                # logging.debug("Position {} pass min_DP filter, with RO {} > {}".format(hgvs, str(RO), str(min_DP)))
                if not SB_range[0] <= SBR <= SB_range[1]:
                    logging.debug("Position {} failed SBR filter {}".format(hgvs, str(SBR)))
                    samp_SBR_fil.append(0)

                if MQMR < min_MQMR:
                    samp_MQMR_fil.append(0)

                if AQR < min_AQR:
                    samp_AQR_fil.append(0)

            if AO > min_DP:
                # logging.debug("Position {} pass min_DP filter, with AO {} > {}".format(hgvs, str(AO), str(min_DP)))
                if not SB_range[0] <= SBA <= SB_range[1]:
                    logging.debug("Position {} failed SBA filter {}".format(hgvs, str(round(SBA, 3))))
                    samp_SBA_fil.append(0)

            if POS in blacklist:
                samp_POS_fil.append(0)

        # if they aren't all zero than at least one passed
        # only one sample has to pass
        FILTER = []
        if len(samp_SBR_fil) < len(FORMAT):
            SBR_fil = 1
        else:
            SBR_fil = 0
            FILTER.append("SBR")

        if len(samp_SBA_fil) < len(FORMAT):
            SBA_fil = 1
        else:
            SBA_fil = 0
            FILTER.append("SBA")

        if len(samp_MQMR_fil) < len(FORMAT):
            MQMR_fil = 1
        else:
            MQMR_fil = 0
            FILTER.append("MQMR")

        if len(samp_AQR_fil) < len(FORMAT):
            AQR_fil = 1
        else:
            AQR_fil = 0
            FILTER.append("AQR")

        if len(samp_POS_fil) < len(FORMAT):
            POS_fil = 1
        else:
            POS_fil = 0
            FILTER.append("POS")

        # FILTER = "FAIL"
        if sum([SBR_fil, SBA_fil, MQMR_fil, AQR_fil, POS_fil]) == 5:
            FILTER = "PASS"
        else:
            FILTER = ";".join(FILTER)
        # print(FILTER)

        line[6] = FILTER

        # line.append(FILTER_INFO)
        # print(line)
        # if POS == 302:
        #   print(info)
        #   print(POS_fil)
        #   print(line)
        new_vcf.append(line)
    # print(new_vcf)
    return (new_vcf)


def update_header(col_names, header_lines, p, SB_range=[0.1, 0.9], min_MQMR=30, min_AQR=20):
    ##############################
    ####### Process Header lines
    header_lines.append([
        '##ALT=<ID=NON_REF,Description="Represents any '
        'possible alternative allele at this location">'])
    ##############################
    ######## INFO lines
    # DP - across all the samples. Is this then summed across all the samples?
    header_lines.append([
        '##INFO=<ID=DP,Number=1,Type=Integer,'
        'Description="Approximate read depth; some reads may '
        'have been filtered">'])
    # MQM - according to freebayes there should be one per allele (Number=A).
    header_lines.append([
        '##INFO=<ID=MQM,Number=A,Type=Float,Description="Mean '
        'mapping quality of observed alternate alleles">'])
    # MRMR only ever one. Could also add to format field, see MQM
    header_lines.append([
        '##INFO=<ID=MQMR,Number=1,Type=Float,'
        'Description="Mean mapping quality of observed '
        'reference alleles">'])
    # QA Again one per allele. But there are sample specific alleles in the
    # format field
    header_lines.append([
        '##INFO=<ID=QA,Number=A,Type=Integer,'
        'Description="Alternate allele quality sum in phred">'])
    # QR only one per reference. Sample specific in the format field
    header_lines.append([
        '##INFO=<ID=QR,Number=1,Type=Integer,'
        'Description="Reference allele quality sum in phred">'])
    # SAF, SAR, SRF, SRR - onle per allele, averaged over all samples. Consider
    # adding sample specific SBR and SBA
    header_lines.append([
        '##INFO=<ID=SAF,Number=A,Type=Integer,'
        'Description="Total number of alternate observations '
        'on the forward strand">'])
    header_lines.append([
        '##INFO=<ID=SAR,Number=A,Type=Integer,'
        'Description="Total number of alternate observations '
        'on the reverse strand">'])
    header_lines.append([
        '##INFO=<ID=SBA,Number=1,Type=Float,'
        'Description="Strand bias of the alternate reads, '
        'SBA=SAF/(SAF+SAR)">'])
    header_lines.append([
        '##INFO=<ID=SBR,Number=1,Type=Float,'
        'Description="Strand bias of the reference reads, '
        'SBR=SRF/(SRF+SRR)">'])
    header_lines.append([
        '##INFO=<ID=SRF,Number=1,Type=Integer,'
        'Description="Number of reference observations on the '
        'forward strand">'])
    header_lines.append([
        '##INFO=<ID=SRR,Number=1,Type=Integer,'
        'Description="Number of reference observations on the '
        'reverse strand">'])
    header_lines.append([
        '##INFO=<ID=TYPE,Number=1,Type=String,'
        'Description="normalised := when the variant has been normalised, complex := when the new reference and alternate length are still > 1 after normalising">'])
    # vaf - will be one for each allele. it is very useful so its good to be in
    # the INFO field
    ##############################
    ######## FILTER lines:
    header_lines.append([
        '##FILTER=<ID=POS,Description="Variant falls in '
        'the blacklist of positions: MT:302-319, '
        'MT:3105-3108">'])
    header_lines.append([
        '##FILTER=<ID=SBR,Description="For all alleles RO '
        '> 15 and (SBR > ' + str(SB_range[1]) + ' or SBR < ' + str(SB_range[0]) + ')">'])
    header_lines.append([
        '##FILTER=<ID=SBA,Description="For all alleles AO '
        '> 15 and (SBA > ' + str(SB_range[1]) + ' or SBA < ' + str(SB_range[0]) + ')">'])
    header_lines.append(
        ['##FILTER=<ID=MQMR,Description="For all alleles MQMR<' + str(min_MQMR) + '">'])
    header_lines.append(
        ['##FILTER=<ID=AQR,Description="For all alleles AQR<' + str(min_AQR) + '">'])
    # header_lines.append(['##FILTER=<ID=VAF,Number=A,Type=Float,
    # Description="Allele frequency in the range (0,1] - the ratio of the
    # number of alternate reads to reference reads">'])
    ##############################
    ######## FORMAT lines
    # GT, DP, RO, AO, QR, QA, AD
    header_lines.append([
        '##FORMAT=<ID=AD,Number=.,Type=Integer,'
        'Description="Allelic depths for the ref and alt '
        'alleles in the order listed">'])
    ##FORMAT=<ID=AD,Number=.,Type=Integer,Description="Allelic depths for the
    # ref and alt alleles in the order listed">
    header_lines.append([
        '##FORMAT=<ID=AO,Number=1,Type=String,'
        'Description="Alternate observations">'])
    header_lines.append([
        '##FORMAT=<ID=AQA,Number=A,Type=Float,'
        'Description="Average base quality of the alternate '
        'reads, AQA=QA/AO">'])
    header_lines.append([
        '##FORMAT=<ID=AQR,Number=A,Type=Float,'
        'Description="Average base quality of the reference '
        'reads, AQR=QR/RO">'])
    header_lines.append([
        '##FORMAT=<ID=DP,Number=1,Type=Integer,'
        'Description="Read depth, AO+RO">'])
    header_lines.append([
        '##FORMAT=<ID=GT,Number=1,Type=String,'
        'Description="Genotype. Variants with less than 4 alt reads are 0/0, '
        'then variants with VAF < 0.95 are 0/1, and variants with VAF >= 0.95 are 1/1">'])
    header_lines.append([
        '##FORMAT=<ID=QA,Number=A,Type=Integer,Description="Alternate allele quality sum in phred">'])
    header_lines.append([
        '##FORMAT=<ID=QR,Number=1,Type=Integer,Description="Reference allele quality sum in phred">'])
    header_lines.append([
        '##FORMAT=<ID=RO,Number=1,Type=String,Description="Reference observations">'])
    header_lines.append([
        '##FORMAT=<ID=VAF,Number=A,Type=Float,Description="Allele frequency in the range (0,1] - the ratio of the number of alternate reads to reference reads">'])
    header_lines.append([
        '##FORMAT=<ID=q,Number=A,Type=Float,Description="Phred scaled binomial probability of seeing AO reads from DP, assuming a noise floor of p=' + str(
            p) + '. ">'])

    ##############################
    ######## Add col names back in
    header_lines.append([col_names])


def do_normalise(vcf, out_file=None, p=0.002, SB_range=[0.1,0.9], min_MQMR=30, min_AQR=20, chromosome=None, genome="mitylib/reference/hs37d5.genome"):
    """
    Normalise and FILTER a mity VCF file.

    This function splits multi-allelic variants and MNPs, whilst taking care to
    split the VCF header properly. After normalising, it also updates the FILTER,
    to remove poor quality variants, and calculates QUAL and GENOTYPE:QUAL scores.

    :param vcf: the path to a vcf.gz file. It can be single-sample,
    or multi-sample, and
         should have been generated by mity call, or FreeBayes
    :type vcf: str

    :param out_file: The name of the normalised vcf.gz file
    :type out_file: str

    :param p: The minimum noise threshold.
    :type p: float

    :param chromosome: discard variants not on this chromosome. If None,
    then all variants will be processed.
    :type chromosome: str

    :returns: Nothing. This creates a vcf.gz named out_file
    :rtype: None
    """
    if out_file is None:
        out_file = vcf.replace(".vcf.gz", ".norm.vcf.gz")

    file = gzip.open(vcf, 'rt')

    # split the header and the variants into two seperate lists
    # @TODO: refactor this to use pyvcf
    logging.debug('Splitting header and variants')
    col_names = None
    header_lines = []
    variants = []
    for line in file:
        if line[0] == "#":
            if re.match("#CHROM", line):
                col_names = line.strip()
                continue
            if re.match("##FORMAT", line) or re.match("##INFO", line):
                continue
            # only keep the non-FORMAT and non-INFO lines. We'll add these back later.
            line = line.strip()
            header_lines.append([line])
        else:
            line = line.strip()
            line_chromosome = line.split('\t')[0]
            if chromosome is None or line_chromosome == chromosome:
                variants.append(line)

    if len(variants) == 0:
        logging.warning('No variants in VCF with specified chromosome/s')

    logging.debug('Splitting multiallelic')
    single_allele = split_multi_allelic(variants)

    logging.debug('Splitting MNPs')
    no_mnp = split_MNP(single_allele)
    # debug_print_vcf_lines(no_mnp)
    logging.debug('Combining duplicated variants')
    combined_variants = combine_lines(no_mnp, p=p)
    # debug_print_vcf_lines(combined_variants)
    logging.debug('Adding filter variants')
    sba_filter = []
    filtered_variants = add_filter(combined_variants, SB_range=SB_range, min_MQMR=min_MQMR, min_AQR=min_AQR)
    # debug_print_vcf_lines(filtered_variants)
    update_header(col_names, header_lines, p=p, SB_range=SB_range, min_MQMR=min_MQMR, min_AQR=min_AQR)
    # debug_print_vcf_lines(filtered_variants)
    logging.info('Writing normalised vcf: {}'.format(out_file))

    new_vcf = header_lines + filtered_variants

    write_vcf(new_vcf, out_file, genome)


def debug_print_vcf_lines(x):
    for line in x:
        print(*line, sep='\t')
    sys.exit()



"""Mitochondrial variant calling."""
import sys
import subprocess
import logging
import os.path
from .util import tabix, check_missing_file, create_prefix, bam_get_mt_contig, bam_has_RG
from .normalise import do_normalise as vcfnorm

def do_call(bam_files, reference, genome=None, prefix=None, min_mq=30, min_bq=24,
            min_af=0.01, min_ac=4, p=0.002, normalise=True, 
            out_folder_path=".", region=None):
    """
    Run mity call.
    :param bam_files: a list of bam_files
    :param reference: the path to the reference genome file (fasta format)
    :param genome: the path to the reference genome file for gsort (genome format). Required if normalise=True
    :param prefix: The result filename prefix. If None, then the first bam_file prefix
    will be used
    :param min_mq: minimum mapping quality threshold. default: 30
    :param min_bq: minimum base quality threshold. default: 24
    :param min_af: minimum heteroplasmy, aka the minimum fraction of alt reads vs total reads.
    scale [0,1]; default 0.005
    :param min_ac: minimum number of alternative reads to support a variant. default: 4
    :param p: the noise threshold. default 0.002
    :param normalise:
    :param out_folder_path: the folder to store the results within. default: .
    :param region: Which region to analyse? If None, then the whole MT will be analysed.
    :return:
    """
    bam_files = bam_files[0] 
    #####
    # Checks
    #####

    if len(bam_files) > 1 and prefix is None:
        raise ValueError(
                "If there is more than one bam file, --prefix must be set")
    
    check_missing_file(bam_files, die=True)
    prefix = create_prefix(bam_files[0], prefix)

    if not all(map(bam_has_RG, bam_files)):
        logging.error("At least one BAM file lacks an @RG header")
        exit(1)

    if normalise and genome is None:
        logging.error("A genome file should be supplied if mity call normalize=True")
        exit(1)

    if not os.path.exists(out_folder_path):
        os.makedirs(out_folder_path)

    output_file_name = os.path.join(out_folder_path, prefix + ".mity.vcf.gz")
    unnormalised_vcf_path = os.path.join(out_folder_path, prefix + ".unnormalised.vcf.gz")

    # if given BAM files A, B, C; freebayes will put samples C, B, A in the VCF file; reverse this odd behaviour
    bam_str = " ".join(['-b ' + bam_file for bam_file in reversed(bam_files)])

    if region is None:
        region = bam_get_mt_contig(bam_files[0], as_string=True)

    # embed the mity command into the VCF header
    mity_cmd = '##mityCommandline="mity call --reference ' + str(reference) + \
               ' --prefix ' + prefix + ' --min-mapping-quality ' + str(min_mq) + \
               ' --min-base-quality ' + str(min_bq) + ' --min-alternate-fraction ' + \
               str(min_af) + ' --min-alternate-count ' + str(min_ac) + \
               ' --out-folder-path ' + str(out_folder_path) + ' --region ' + region
    logging.debug("mity commandline: " + str(mity_cmd))

    if normalise:
        mity_cmd = mity_cmd + ' --normalise --p ' + str(p)

    mity_cmd = mity_cmd + ' ' + ' '.join(bam_files)

    mity_cmd = mity_cmd + '"'
    mity_cmd = mity_cmd.replace("/", "\/")
    logging.debug(mity_cmd)

    # overwrite a redundant freebayes header line with the mity command line
    sed_cmd = "sed 's/^##phasing=none/{}/g'".format(mity_cmd)
    logging.debug(sed_cmd)

    freebayes_call = ('set -o pipefail && freebayes -f {} {} '
                      '--min-mapping-quality {} '
                      '--min-base-quality {} '
                      '--min-alternate-fraction {} '
                      '--min-alternate-count {} '
                      '--ploidy 2 '
                      '--region {} '
                      ).format(reference, bam_str, min_mq, min_bq, min_af, min_ac, region)
    freebayes_call = freebayes_call + ('| sed "s/##source/##freebayesSource/" | sed "s/##commandline/##freebayesCommandline/" | {} | bgzip > {} ').format(sed_cmd, unnormalised_vcf_path)

    logging.info("Running FreeBayes in sensitive mode")
    logging.debug(freebayes_call)
    res = subprocess.run(freebayes_call, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable='/bin/bash')
    logging.debug("Freebayes result code: {}".format(res.returncode))

    if res.returncode != 0:
        logging.error("FreeBayes failed: {}".format(res.stderr))
        exit(1)

    if os.path.isfile(unnormalised_vcf_path):
        logging.debug("Finished running FreeBayes")
    
    if normalise:
        logging.debug("Normalising and Filtering variants")
        try:
            vcfnorm(vcf=unnormalised_vcf_path, out_file=output_file_name, p=p, genome=genome)
        finally:
            os.remove(unnormalised_vcf_path)
    else:
        os.rename(unnormalised_vcf_path, output_file_name)
        tabix(output_file_name)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("mitylib/_version.py", "r") as fh:
    version = fh.read().replace("__version__ = ", "").strip('""\n')

setuptools.setup(
    name="mitywgs",
    version=version,
    description="A sensitive Mitochondrial variant detection pipeline from WGS data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KCCG/mity",
    author="Clare Puttick",
    author_email="clare.puttick@gmail.com",
    license="MIT",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'License :: Free for non-commercial use',
        'Natural Language :: English',
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    keywords='mitochondrial DNA genomics variant SNV INDEL',
    project_urls={
        'Documentation': 'https://github.com/KCCG/mity/',
        'Source': 'https://github.com/KCCG/mity/',
        'Funding': 'http://garvan.org.au/kccg',
    },
    packages=setuptools.find_packages(),
    install_requires=[
        'pyvcf',
        'pysam',
        'pandas',
        'xlsxwriter',
        'pyfastx',
        'scipy'
    ],
    python_requires='>=3.5.4',
    package_data={'mitylib': ['verchew.ini', 'annot/*', 'reference/*']},
    include_package_data=True,
    scripts=["mity"]
)



import sys
import logging
import gzip
import pandas
import os.path
import xlsxwriter
from .util import check_missing_file, create_prefix, make_hgvs, get_annot_file

def make_table(variants, samples, vep_headers, impact_dict, min_vaf):
    table = []
    n_sample = len(samples)
    
    for lineparts in variants:
        # These should be the same for all VCFs
        chromosome = lineparts[0]
        pos = lineparts[1]
        ref = lineparts[3]
        alt = lineparts[4]
        QUAL = lineparts[5]
        FILTER = lineparts[6]

        hgvs = make_hgvs(pos, ref, alt)
        # print(hgvs)
        info = lineparts[7]
        info = info.split(";")
        # print(info)
        
        ## get the filters
        
        POS_FILTER = "FAIL" if 'POS' in FILTER else "PASS"
        SBR_FILTER = "FAIL" if 'SBR' in FILTER else "PASS"
        SBA_FILTER = "FAIL" if 'SBA' in FILTER else "PASS"
        MQMR_FILTER = "FAIL" if 'MQMR' in FILTER else "PASS"
        AQR_FILTER = "FAIL" if 'AQR' in FILTER else "PASS"

        INFO_names = [x.split("=")[0] for x in info]
        INFO_values = [i.split("=")[1] for i in info]
        # print(INFO_names)
        # print(INFO_values)
        
        ############################################
        ########## VEP
        if vep_headers != "":
            
            VEP_idx = INFO_names.index('CSQ')
            VEP = INFO_values[VEP_idx]
            VEP = VEP.split(",")
            VEP = [v.split('|') for v in VEP]
            # print(VEP)
            
            # get the index of where the consequence sits in the list of VEP 
            # annotations
            consequence_idx = vep_headers.index('Consequence')
            
            # get the index of the annotations that don't contain the word stream
            not_stream = ['stream' not in x[consequence_idx] for x in VEP]
            stream_idx = [i for i, x in enumerate(not_stream) if x]
            # print(stream_idx)
            
            # remove the sublists that have 'stream'in the consequence from 
            # the VEP annotations
            VEP = [VEP[i] for i in stream_idx]
            # print(VEP)
            # remove empty sublists
            VEP = [x for x in VEP if x != []]
            # print(VEP)
            
            # if there are no annotations left, completely remove the sublist
            if VEP == []:
                # fill with dot
                VEP = [['.'] * len(vep_headers)]
            else:
                VEP = [[element or '.' for element in sublist] for sublist in
                       VEP]
            
            # print(VEP)
            
            # get all the consequences for each variant in a single list
            # so that we can see what the highest impact is
            csq = [x[consequence_idx] for x in VEP]
            # some of them are split by '&'
            csq = [x.split("&") for x in csq]
            # unlist
            csq = sum(csq, [])
            # print(csq)
            
            # print(impact)
            if csq == ['.']:
                # impact = ["."]
                highest_impact = "."
            else:
                impact = [impact_dict[x] for x in csq]
                if 'HIGH' in impact:
                    highest_impact = 'HIGH'
                elif 'MODERATE' in impact:
                    highest_impact = 'MODERATE'
                elif 'MODIFIER' in impact:
                    highest_impact = 'MODIFIER'
                elif 'LOW' in impact:
                    highest_impact = 'LOW'
                else:
                    sys.error("unknown SO term")
                    
                    # print(highest_impact)
            
            # concatenate the vep annotations together if there is more than 
            # one annotation per variant
            vep_annotations = []
            for ann in range(0, len(vep_headers)):
                t = [x[ann] for x in VEP]
                # print(t)
                t = ";".join(t)
                vep_annotations.append(t)
            vep_annotations = vep_annotations + [highest_impact]
            # print("")
        # print(vep_annotations)
        # print("") 
        
        # print(vep_headers)
        # print(sublist_length)
        # print(VEP)
        
        ############################################
        ########## INFO
        
        DP_idx = INFO_names.index('DP')
        INFO_DP = (INFO_values[DP_idx])
        
        MQM_idx = INFO_names.index('MQM')
        INFO_MQM = (INFO_values[MQM_idx])
        
        MQMR_idx = INFO_names.index('MQMR')
        INFO_MQMR = (INFO_values[MQMR_idx])
        
        QA_idx = INFO_names.index('QA')
        INFO_QA = (INFO_values[QA_idx])
        
        QR_idx = INFO_names.index('QR')
        INFO_QR = (INFO_values[QR_idx])
        
        SAF_idx = INFO_names.index('SAF')
        INFO_SAF = (INFO_values[SAF_idx])
        
        SAR_idx = INFO_names.index('SAR')
        INFO_SAR = (INFO_values[SAR_idx])
        
        SRF_idx = INFO_names.index('SRF')
        INFO_SRF = (INFO_values[SRF_idx])
        
        SRR_idx = INFO_names.index('SRR')
        INFO_SRR = (INFO_values[SRR_idx])
        
        SBR_idx = INFO_names.index('SBR')
        INFO_SBR = (INFO_values[SBR_idx])
        
        SBA_idx = INFO_names.index('SBA')
        INFO_SBA = (INFO_values[SBA_idx])
        
        # VAF_idx = INFO_names.index('VAF')
        # INFO_VAF = (INFO_values[VAF_idx])
        # INFO_VAF = INFO_VAF.split(",")
        # print(VAF)
        
        geno_names = lineparts[8]
        geno_names = geno_names.split(":")
        # print(geno_names)
        
        for samp in range(0, n_sample):
            
            # samp_VAF = INFO_VAF[samp]
            # print(samp_VAF)
            
            geno = lineparts[9 + samp]
            geno = geno.split(":")
            # print(geno)
            
            VAF_idx = geno_names.index('VAF')
            FORMAT_VAF = geno[VAF_idx]
            
            if float(FORMAT_VAF) > float(min_vaf):
                
                # print("VAF")
                # print(float(FORMAT_VAF))
                # print(min_vaf)
                # print(float(FORMAT_VAF) > float(min_vaf))
                
                GT_idx = geno_names.index('GT')
                FORMAT_GT = geno[GT_idx]
                # print(GT)
                
                DP_idx = geno_names.index('DP')
                FORMAT_DP = geno[DP_idx]
                # print(FORMAT_DP)
                
                # AD_idx = geno_names.index('AD')
                # FORMAT_AD = str(geno[AD_idx])
                # print(FORMAT_AD)  
                
                RO_idx = geno_names.index('RO')
                FORMAT_RO = geno[RO_idx]
                # print(FORMAT_RO)  
                
                QR_idx = geno_names.index('QR')
                FORMAT_QR = geno[QR_idx]
                # print(FORMAT_QR)  
                
                AQR_idx = geno_names.index('AQR')
                FORMAT_AQR = geno[AQR_idx]
                # print(FORMAT_AQR) 
                
                AO_idx = geno_names.index('AO')
                FORMAT_AO = geno[AO_idx]
                # print(FORMAT_AO)      
                
                QA_idx = geno_names.index('QA')
                FORMAT_QA = geno[QA_idx]
                # print(FORMAT_QA)
                
                AQA_idx = geno_names.index('AQA')
                FORMAT_AQA = geno[AQA_idx]
                # print(FORMAT_AQA)                                          

                # referred to as variant_quality in the final table
                QUAL_idx = geno_names.index('q')
                FORMAT_QUAL = geno[QUAL_idx]
                # print(FORMAT_QUAL)

                # @TODO - move to `mity normalise` as INFO:TIER
                if float(FORMAT_VAF) >= 0.01:
                    tier = 1
                elif float(FORMAT_VAF) < 0.01 and float(FORMAT_AO) > 10:
                    tier = 2
                else:
                    tier = 3
                    
                # table = [["SAMPLE", "CHR", "POS", "REF", "ALT", "QUAL",
                # "FILTER", "INFO_DP", "INFO_MQM", "INFO_MQMR",
                #       "INFO_QA", "INFO_QR", "INFO_SAF", "INFO_SAR",
                #       "INFO_SRF", "INFO_SRR", "INFO_SBR", "INFO_SBA",
                #       "INFO_VAF", "INFO_POS_FILTER", "INFO_SBR_FILTER",
                #       "INFO_SBA_FILTER", "INFO_MQMR_FILTER", "INFO_AQR_FILTER",
                #       "FORMAT_GT", "FORMAT_DP", "FORMAT_RO",
                #       "FORMAT_QR", "FORMAT_AQR", "FORMAT_AO",
                #       "FORMAT_QA", "FORMAT_AQA", "INFO", "FORMAT"]]
                # print(lineparts[7])
                # print(lineparts[9+samp])
                
                no_comma_info = str(lineparts[7])
                no_comma_info = no_comma_info.replace(",", "_")
                
                no_comma_format = str(lineparts[9 + samp])
                no_comma_format = no_comma_format.replace(",", "_")
                
                new_line = [samples[samp], chromosome, pos, ref, alt, hgvs,
                            QUAL, FILTER, INFO_DP, INFO_MQM, INFO_MQMR,
                            INFO_QA, INFO_QR, INFO_SAF, INFO_SAR, INFO_SRF,
                            INFO_SRR, INFO_SBR, INFO_SBA, FORMAT_VAF,
                            POS_FILTER, SBR_FILTER, SBA_FILTER, MQMR_FILTER, AQR_FILTER,
                            FORMAT_GT, FORMAT_DP, FORMAT_RO, FORMAT_QR,
                            FORMAT_AQR, FORMAT_AO,
                            FORMAT_QA, FORMAT_AQA, FORMAT_QUAL, tier, no_comma_info,
                            no_comma_format]
                
                if vep_headers != "":
                    new_line = new_line + vep_annotations

                table.append(new_line)
    
    return (table)

def split_header_variants(vcf):
    header = []
    variants = []
    for line in vcf:
        # print(line)
        if line[0] == '#':
            line = line.strip()
            header.append(line)
            # print(line)
        else:
            line = line.strip()
            # split the line into its columns
            lineparts = line.split('\t')
            variants.append(lineparts)
            
            # print(header)
    # print(variants)
    
    # check if VEP appears anywhere in the header
    is_vepped = ['VEP' in x for x in header]
    is_vepped = True in is_vepped
    # print(is_vepped)
    
    # if its vepped we need to get the vep col names
    if is_vepped:
        csq_line_idx = ['INFO=<ID=CSQ' in x for x in header]
        csq_line_idx = csq_line_idx.index(True)
        csq_line = header[csq_line_idx]
        csq_line = csq_line.strip()
        # remove the last "> from the line)
        vep_header = csq_line[:-2]
        vep_header = vep_header.split("|")
        # remove the first uneeded stuff before Consequence
        vep_header[0] = vep_header[0].rsplit(' ', 1)[1]
    
    
    else:
        vep_header = ""
        # impacts = ""      
        # print(d)
        # print(d['non_coding_transcript_exon_variant'])
    
    # the col names should be the last line of the header
    col_names = header[-1]
    # print(col_names)
    # check that the col names only start with a #
    if not (col_names[0] == '#' and col_names[1] != '#'):
        sys.exit(
            "It looks like the last line of the header is not the column "
            "names.")
    col_names = col_names.strip()
    # remove the first # from CHROM
    col_names = col_names[1:]
    col_names = col_names.split('\t')
    # print(col_names)
    samples = col_names[9:]
    n_sample = len(samples)
    # remove the last n_sample items from the col_names
    del col_names[-n_sample:]
    return (samples, col_names, variants, is_vepped, vep_header, header)


def find_index(string, pattern):
    return [i for i, ltr in enumerate(string) if ltr == pattern]


def do_report(vcf, prefix=None, min_vaf=0.0, out_folder_path = "."):
    """
    Create a mity report
    :param vcf: the path to a vcf file
    :param prefix: the optional prefix. This must be set if there is >1 vcf files
    :param min_vaf: only include vairants with vaf > min_vaf in the report
    :return:
    """
    vcf = vcf[0]
    
    if len(vcf) == 0:
        raise ValueError("At least one VCF file must be supplied")
    if prefix is None and len(vcf) > 1:
        raise ValueError("If there is more than one vcf file, --prefix must be set")
    check_missing_file(vcf, die=True)

    prefix = create_prefix(vcf[0], prefix)
    
    # loop over all the files that are input
    variant_list = []
    
    samples = []
    col_names = []
    variants = []
    is_vepped = []
    vep_header = []
    header_text = []
    
    for filepath in vcf:
        # this assumes they all have the same VEP cols
        file = gzip.open(filepath, 'rt')
        # for each file get the first line that doesnt start with # to get 
        # the sample names
        
        header_variants = split_header_variants(file)
        
        # there will be multiple sample names per vcf (if joint called)
        # each set of samples a sublist. One vcf has a set of samples
        samples.append(header_variants[0])
        
        # one set of col names per vcf
        # each set of column names a sublist
        file_col_names = header_variants[1]
        col_names.append(file_col_names)
        
        # one set of variants per vcf
        # each set of variants a sublist
        file_variants = header_variants[2]
        variants.append(file_variants)
        
        # each vcf is either vepped or not
        # list of true false for "if vepped"
        is_vepped = is_vepped + [header_variants[3]]
        vep_header.append(header_variants[4])
        
        header_text = header_text + header_variants[5]
    # print(header_text)
    # for the Documentation sheet in the output, get the VCF headers
    # we want the header that explains INFO, FILTER or FORMAT fields
    line_types = ["##INFO", "##FILTER", "##FORMAT"]
    vcf_documentation = []
    ID = []
    for line in header_text:
        line_type = line.split("=")[0]
        if line_type in line_types:
            # print(line)
            ##INFO=<ID=DP,Number=1,Type=Integer,Description="
            description_ID = line.split(",")[0]
            ID.append(description_ID)
            description_ID = description_ID.split("ID=")[1]
            # the CSQ header will be taken care of in the VEP columns
            if description_ID != "CSQ":
                line_type = line_type.split("##")[1]
                if line_type != "FILTER":
                    description_ID = description_ID + "_" + line_type + ": "
                else:
                    description_ID = description_ID + ": "
                description = line.split('Description="')[1]
                description = description[:-2]
                new_doc_line = description_ID + description
                vcf_documentation.append(new_doc_line)
    
    noDupes_vcf_documentation = []
    
    for line in vcf_documentation:
        if not noDupes_vcf_documentation.count(line):
            noDupes_vcf_documentation.append(line)
    # print(noDupes_vcf_documentation)
    # if the number of unique IDs does not match the number of unique 
    # descriptions, then there is one ID 
    # that has more than one descrption, either because of one vcf with 
    # duplicated headers, or because
    # 2 VCFs have been inputted with conflicting headers
    dupe_header_warning = False
    if (len(set(ID)) != len(noDupes_vcf_documentation)):
        dupe_header_warning = True
    # print(dupe_header_warning)
    # sys.exit()    
    
    # check they all have the same length
    # print(samples)
    if len(samples) != len(col_names) != len(variants) != len(is_vepped) != len(
            vep_header):
        sys.exit(
            "length of samples, col_names, variants, is_vepped, vep_header is "
            "different, something odd has happened.")
    
    # check that all the col_names are the same
    if len(col_names) > 1:
        l = col_names.pop(0)
        all_equal = [x == l for x in col_names]
        
        if set(all_equal) != set([True]):
            sys.exit(
                "VCF inputs do not have the same column name for the variants")
    
    # if all the col_names are the same we only need to keep one of them
    # print(col_names)
    col_names = col_names[0]
    # print(col_names)
    
    # check they are either all vepped or not vepped
    if len(set(is_vepped)) != 1:
        sys.exit("Some VCFs are vepped and some are not. They must be the same")
    # again, now all we need to know is if they are all vepped or not
    is_vepped = True in is_vepped
    
    # check that the vep headers are the same
    if len(vep_header) > 1:
        l = vep_header.pop(0)
        all_equal = [x == l for x in vep_header]
        
        if set(all_equal) != set([True]):
            sys.exit("VCFs have different VEP columns")
            # if all the vep_header are the same we only need to keep one of 
            # them
    vep_header = vep_header[0]
    
    # if its vepped we need the impacts
    if is_vepped:
        impacts = {}
        # http://asia.ensembl.org/info/genome/variation/predicted_data.html
        with open("ensemble_impacts.txt") as f:
            for line in f:
                (key, val) = line.split()
                impacts[key] = val
    else:
        impacts = ""
    
    # range!!!!!
    
    ## get the variants into a list format
    variant_table = []
    # print(len(variants))
    # print("making table")
    for vcf_num in range(0, len(variants)):
        v = make_table(variants[vcf_num], samples[vcf_num], vep_header, impacts,
                       min_vaf)
        # variant_table.append(v)w
        variant_table = variant_table + v
    # To print just the variants:
    # for line in variant_table:
    # print(line)
    # sys.exit()    
    
    # note that this is hard coded and the function make_table outputs the 
    # columns in this order
    # the column order is changed later in the code
    vcf_headers = ["SAMPLE", "CHR", "POS", "REF", "ALT", "HGVS", "QUAL",
                   "FILTER", "total_locus_depth", "MQM_INFO", "MQMR_INFO",
                   "QA_INFO", "QR_INFO", "SAF_INFO", "SAR_INFO", "SRF_INFO",
                   "SRR_INFO", "SBR_INFO", "SBA_INFO",
                   "variant_heteroplasmy", "POS_FILTER", "SBR_FILTER", "SBA_FILTER",
                   "MQMR_FILTER", "AQR_FILTER",
                   "GT_FORMAT", "total_sample_depth", "ref_depth", "QR_FORMAT",
                   "AQR_FORMAT", "alt_depth",
                   "QA_FORMAT", "AQA_FORMAT", "variant_quality", "tier", "INFO", "FORMAT"]
    
    if vep_header != "":
        table_vep_headers = [x + "_VEP" for x in vep_header]
        # we add 'highest_vep_impact' to the end of the VEP cols
        vcf_headers = vcf_headers + table_vep_headers
        vcf_headers = vcf_headers + ['highest_vep_impact']
        # print(headers)
    # print(headers)
    
    variant_df = pandas.DataFrame(variant_table, columns=vcf_headers)
    # print(variant_df)
    # variant_df.to_csv("test_variants.csv", index = False)

    # The annotations use contig name = 'MT', so if the vcf uses contig name = 'chrM', then swap this.
    contig_name = variant_df['CHR'][0]
    if contig_name != "MT":
        variant_df['CHR'] = "MT"
    ########## Merge with mitomap and panel annotation table
    # this depends on chrom, pos, ref, alt
    mitomap_panel_annotations = pandas.read_csv(get_annot_file("mitomap_panel_annotations.csv"))
    mitomap_panel_annotations['POS'] = mitomap_panel_annotations['POS'].astype('str')
    # print(mitomap_panel_annotations.dtypes)
    # print(variant_df.dtypes)
    mitomap_panel_annotated_variants = pandas.merge(left=variant_df,
                                                    right=mitomap_panel_annotations,
                                                    how='left',
                                                    on=['CHR', 'POS', 'REF', 'ALT'])
    
    ########## Merge with gene names and biotype
    # this depends on chrom, pos
    gtf_annotations = pandas.read_csv(get_annot_file("gtf_annotations.csv"))
    gtf_annotations['POS'] = gtf_annotations['POS'].astype('str')
    # print(gtf_annotations.dtypes)
    
    gtf_mitomap_panel_annotated_variants = pandas.merge(
        left=mitomap_panel_annotated_variants, right=gtf_annotations,
        how='left', on=['CHR', 'POS'])
    # annotated_variants = annotated_variants.fillna('.')
    
    ########## Merge with Mitomap Locations
    # this depends on chrom, pos
    mito_locus_annotations = pandas.read_csv(get_annot_file("mito_dna_func_loc.csv"))
    mito_locus_annotations['POS'] = mito_locus_annotations['POS'].astype('str')
    # print(gtf_annotations.dtypes)
    
    mito_locus_gtf_mitomap_panel_annotated_variants = pandas.merge(
        left=gtf_mitomap_panel_annotated_variants, right=mito_locus_annotations,
        how='left', on=['CHR', 'POS'])
    # for line in mito_locus_gtf_mitomap_panel_annotated_variants:
    #   print(line)
    # sys.exit()
    
    ########## Merge with trna anticodon positions
    # this depends on chrom pos 
    trna_annotations = pandas.read_csv(get_annot_file("anticodon_positions.csv"))
    trna_annotations['POS'] = trna_annotations['POS'].astype('str')
    trna_annotations['anticodon'] = trna_annotations['anticodon'].astype('str')
    trna_mito_locus_gtf_mitomap_panel_annotated_variants = pandas.merge(
        left=mito_locus_gtf_mitomap_panel_annotated_variants,
        right=trna_annotations, how='left', on=['CHR', 'POS'])
    # for line in trna_mito_locus_gtf_mitomap_panel_annotated_variants:
    #   print(line)
    # sys.exit()
    
    ########## Merge with mitotip score and prediction
    # this depends on chrom pos ref alt
    mitotip_annotations = pandas.read_csv(get_annot_file("mitotip_score_fixed_del.csv"))
    mitotip_annotations['POS'] = mitotip_annotations['POS'].astype('str')
    mitotip_annotations['MitoTip_score'] = mitotip_annotations[
        'MitoTip_score'].astype('str')
    mitotip_annotations['MitoTip_percentile'] = mitotip_annotations[
        'MitoTip_percentile'].astype('str')
    mitotip_annotations['MitoTip_interpretation'] = mitotip_annotations[
        'MitoTip_interpretation'].astype('str')
    mitotip_trna_mito_locus_gtf_mitomap_panel_annotated_variants = pandas.merge(
        left=trna_mito_locus_gtf_mitomap_panel_annotated_variants,
        right=mitotip_annotations, how='left', on=['CHR', 'POS', 'REF', 'ALT'])
    # for line in trna_mito_locus_gtf_mitomap_panel_annotated_variants:
    #   print(line)
    # sys.exit()
    
    ########## Merge with mgrb variants
    # this depends on chrom pos ref alt
    mgrb_annotations = pandas.read_csv(get_annot_file("mgrb_variants.csv"))
    mgrb_annotations['POS'] = mgrb_annotations['POS'].astype('str')
    mgrb_mitotip_trna_mito_locus_gtf_mitomap_panel_annotated_variants = \
        pandas.merge(
        left=mitotip_trna_mito_locus_gtf_mitomap_panel_annotated_variants,
        right=mgrb_annotations, how='left', on=['CHR', 'POS', 'REF', 'ALT'])
    # for line in trna_mito_locus_gtf_mitomap_panel_annotated_variants:
    # print(line)
    # sys.exit()
    
    ########## Merge with haplotype data
    # TODO: why does this add two lines for some variants? (there are no pos-ref-alt dups)
    haplotype_annotations = pandas.read_csv(get_annot_file("haplotype_data.csv"))
    haplotype_annotations['POS'] = haplotype_annotations['POS'].astype('str')
    haplotype_mgrb_mitotip_trna_mito_locus_gtf_mitomap_panel_annotated_variants = pandas.merge(
        left=mgrb_mitotip_trna_mito_locus_gtf_mitomap_panel_annotated_variants,
        right=haplotype_annotations, how='left',
        on=['CHR', 'POS', 'REF', 'ALT'])
    # for line in trna_mito_locus_gtf_mitomap_panel_annotated_variants:
    # print(line)
    # sys.exit()
    
    annotated_variants = \
        haplotype_mgrb_mitotip_trna_mito_locus_gtf_mitomap_panel_annotated_variants.fillna(
        '.')
    # print(annotated_variants)
    # sys.exit()

    # The annotations use contig name = 'MT', so if the vcf uses contig name = 'chrM', then swap this.
    if contig_name != "MT":
        annotated_variants["CHR"] = contig_name

    # now we make the general gene/locus and gene/locus description cols
    gene = annotated_variants['GENE'].tolist()
    gene_biotype = annotated_variants['GENE_BIOTYPE'].tolist()
    mitomap_locus = annotated_variants['Map_Locus'].tolist()
    mitomap_description = annotated_variants['Description'].tolist()
    
    general_locus = gene
    general_biotype = gene_biotype
    # now replace "."'s with mitomap entries
    empty_idx = find_index(general_locus, ".")
    for x in empty_idx:
        general_locus[x] = mitomap_locus[x]
        general_biotype[x] = mitomap_description[x]
    
    # we remove gene, gene biotype, Map_Locus and Description from table
    del annotated_variants['GENE']
    del annotated_variants['GENE_BIOTYPE']
    del annotated_variants['Map_Locus']
    del annotated_variants['Description']
    del annotated_variants['baylor_panel']
    del annotated_variants['common_22_panel']
    del annotated_variants['common_58_panel']
    #annotated_variants1.drop(columns=['baylor_panel','common_22_panel','common_58_panel'])

    # add the general locus and general locus description
    annotated_variants['gene/locus'] = general_locus
    annotated_variants['gene/locus description'] = general_biotype
    
    cohort_count_table = pandas.DataFrame(
        {'count': annotated_variants.groupby(["POS", "ALT"]).size()}).reset_index()
    cohort_count_table.columns = ['POS', 'ALT', 'COHORT_COUNT']
    
    annotated_variants1 = pandas.merge(left=annotated_variants,
                                       right=cohort_count_table, how='left',
                                       on=['POS', 'ALT'])
    
    # add in the GB percent
    # need to account for the fact that some entries will be "." and cant be 
    # converted to float
    gb_freq = annotated_variants1['GenBank_frequency_mitomap'].tolist()
    gb_perc = gb_freq
    
    # only convert the entries that are not a "."
    str_idx = find_index(gb_freq, ".")
    str_idx = [i for i, ltr in enumerate(gb_freq) if ltr != "."]
    
    for x in str_idx:
        freq = round(float(gb_freq[x]) / 32059, 3)
        # print(freq)
        gb_perc[x] = freq
    
    annotated_variants1['allele_frequency_mitomap'] = gb_perc
    # TODO: to make this more general, choose some important cols to go first
    # if they are in the header then do this sorting
    # otherwise dont
    if is_vepped:
        cols = ['SAMPLE', 'HGVS', 'gene/locus', 'gene/locus description',
                'variant_heteroplasmy', 'alt_depth',
                'ref_depth', 'total_sample_depth', 'variant_quality',
                'total_locus_depth', 'COHORT_COUNT', 'tier', 'commercial_panels',
                'phylotree_haplotype', 'MitoTip_score',
                'MitoTip_percentile', 'MitoTip_interpretation', 'anticodon',
                'allele_frequency_mitomap',
                'highest_vep_impact', 'Consequence_VEP', 'disease_mitomap',
                'MGRB_frequency', 'MGRB_FILTER', 'MGRB_AC', 'MGRB_AN', 'phylotree_mut',
                'Codons_VEP', 'Amino_acids_VEP', 'Gene_VEP', 'SYMBOL_VEP',
                'Feature_VEP', 'EXON_VEP', 'PolyPhen_VEP', 'SIFT_VEP',
                'Protein_position_VEP', 'BIOTYPE_VEP', 'CANONICAL_VEP',
                'Feature_type_VEP', 'cDNA_position_VEP', 'CDS_position_VEP',
                'Existing_variation_VEP', 'DISTANCE_VEP', 'STRAND_VEP',
                'CLIN_SIG_VEP', 'LoF_flags_VEP', 'LoF_filter_VEP', 'LoF_VEP',
                'RadialSVM_score_VEP', 'RadialSVM_pred_VEP', 'LR_score_VEP',
                'LR_pred_VEP', 'CADD_raw_VEP', 'CADD_phred_VEP',
                'Reliability_index_VEP', 'HGVSc_VEP', 'HGVSp_VEP',
                'locus_mitomap', 'num_references_mitomap',
                'variant_amino_acid_change_mitomap', 'codon_position_mitomap',
                'codon_number_mitomap',
                'num_disease_references_mitomap', 'RNA_mitomap',
                'homoplasmy_mitomap', 'heteroplasmy_mitomap', 'status_mitomap',
                'disease_amino_acid_change_mitomap', 'GenBank_frequency_mitomap',
                'CHR', 'POS', 'REF', 'ALT', 'QUAL', 'FILTER',
                'MQM_INFO', 'MQMR_INFO', 'QA_INFO', 'QR_INFO', 'SAF_INFO',
                'SAR_INFO', 'SRF_INFO', 'SRR_INFO', 'SBR_INFO', 'SBA_INFO',
                'POS_FILTER', 'SBR_FILTER', 'SBA_FILTER', 'MQMR_FILTER', 'AQR_FILTER',
                'GT_FORMAT', 'QR_FORMAT', 'AQR_FORMAT', 'QA_FORMAT',
                'AQA_FORMAT',
                'INFO', 'FORMAT']
    else:
        cols = ['SAMPLE', 'HGVS', 'gene/locus', 'gene/locus description',
                'variant_heteroplasmy', 'alt_depth',
                'ref_depth', 'total_sample_depth', 'variant_quality',
                'total_locus_depth', 'COHORT_COUNT', 'tier', 'commercial_panels',
                'phylotree_haplotype', 'MitoTip_score',
                'MitoTip_percentile', 'MitoTip_interpretation', 'anticodon',
                'allele_frequency_mitomap',
                'disease_mitomap',
                'MGRB_frequency', 'MGRB_FILTER', 'MGRB_AC', 'MGRB_AN',
                'phylotree_mut',
                'locus_mitomap', 'num_references_mitomap',
                'variant_amino_acid_change_mitomap', 'codon_position_mitomap',
                'codon_number_mitomap',
                'num_disease_references_mitomap', 'RNA_mitomap',
                'homoplasmy_mitomap', 'heteroplasmy_mitomap', 'status_mitomap',
                'disease_amino_acid_change_mitomap', 'GenBank_frequency_mitomap',
                'CHR', 'POS', 'REF', 'ALT', 'QUAL', 'FILTER',
                'MQM_INFO', 'MQMR_INFO', 'QA_INFO', 'QR_INFO', 'SAF_INFO',
                'SAR_INFO', 'SRF_INFO', 'SRR_INFO', 'SBR_INFO', 'SBA_INFO',
                'POS_FILTER', 'SBR_FILTER', 'SBA_FILTER', 'MQMR_FILTER', 'AQR_FILTER',
                'GT_FORMAT', 'QR_FORMAT', 'AQR_FORMAT', 'QA_FORMAT',
                'AQA_FORMAT',
                'INFO', 'FORMAT']
        # These should be equal if you haven't missed any cols
    
    if len(annotated_variants1.columns.tolist()) != len(cols):
        print(cols)
        print("")
        print(annotated_variants.columns.tolist())
        print("")
        print("annotated_variants missing" + str(
            set(annotated_variants.columns.tolist()) - set(cols)))
        print("")
        print(
            "cols" + str(set(cols) - set(annotated_variants.columns.tolist())))
        sys.exit("different number of cols")
    
    annotated_variants1 = annotated_variants1[cols]
    
    # make the documentation sheet
    documentation = []
    x = "This sheet contains the variants from"
    for filename in vcf:
        name = filename.split("/")[-1]
        x = x + " " + str(name)
    documentation.append(x)
    documentation.append("Some column definitions:")
    documentation.append(
        "HGVS: Variant annotation using HGVS syntax, which assumes variant is "
        "SNP, insertion or deletion")
    documentation.append(
        "gene/locus: The gene that the variant falls in, or if it is not in a "
        "gene then the mitomap entry, if it exists.")
    documentation.append(
        "variant_heteroplasmy: alt_depth/(ref_depth+alt_depth)")
    documentation.append(
        "variant_quality: phred-scaled variant quality")
    documentation.append(
        "total_sample_depth: alt_depth + ref_depth for the particular sample")
    documentation.append(
        "total_locus_depth: sum of all the total_sample_depth in the vcf file")
    documentation.append(
        "COHORT_COUNT: The number of times that variant (defined by position "
        "and alternate) is seen in all of the input VCFs")
    documentation.append(
        "tier: 1:=VAF>=1%. 2:= VAF<1% and alt_depth>=10, 3:=VAF<1% and "
        "alt_depth<10")
    documentation.append(
        "commercial_panels: This indicates whether a variant is commonly tested "
        "in a number of commercial/academic testing laboratories. "
        "Baylor: https://www.bcm.edu/research/medical-genetics-labs/test_detail.cfm?testcode=2010; "
        "Common22: https://www.vcgs.org.au/order/tests/634; "
        "Common58: https://www.genedx.com/test-catalog/available-tests"
        "/58-confirmed-disease-causing-mtdna-point-mutations-and-deletion"
        "-testing/")
    documentation.append(
        "MGRB: The Medical Genomics Reference Bank (MGRB) "
         "is a cohort of healthy older Ausrtralian's sequenced using "
         "~30x Illumina HiSeq X at the Garvan Institute, Sydney. "
         "See https://www.biorxiv.org/content/10.1101/473348v1. ")
    documentation.append(
        "MGRB_frequency: represents the frequency of this variant in the "
        "MGRB cohort")
    documentation.append(
        "MGRB_FILTER: represents the FILTER status of the variant in the "
        "MGRB cohort using GATK HaplotypeCaller")
    documentation.append(
        "MGRB_AC: the allele count in the MGRB cohort; ie the number of "
        "individuals with this variant. "
        "This is not accurate for the MT, as it assumes each individual is "
        "diploid. Interpret this value as 2x the number of individuals "
        "in the MGRB control cohort.")
    documentation.append(
        "MGRB_AN: the total number of alleles called in the MGRB cohort. "
        "This is not accurate for the MT, as it assumes each individual is "
        "diploid. Interpret this value as 2x the number of individuals "
        "in the MGRB control cohort.")
    documentation.append(
        "phylotree_haplotype: The haplotypes that the variant is known "
        "to contribute to. See http://www.phylotree.org/index.htm")
    documentation.append(
        "MitoTip_score: MitoTip is an in silico tool for predicting the "
        "pathogenicity of novel mitochondrial tRNA variants, developed "
        "by Neal Sondheimer and Sanjay Sonney (PMID: 29227991). "
        "As such, it only scores variants in a tRNA. The score is a "
        "prediction of the pathogenicity of variants in tRNAs, "
        "see https://www.mitomap.org/foswiki/bin/view/MITOMAP/MitoTipInfo. "
        "The highest score is 21.8. It is far from perfect, as m.3243A>G "
        "is scored as likely benign. ")
    documentation.append(
        "MitoTip_percentile: Where this variants MitoTip score sits in "
        "relation to all of the possible variants in tRNA. 100% is the most "
        "pathogenic, 0% is a score of zero.")
    documentation.append(
        "MitoTip_interpretation: Variants in the upper 25th percentile are "
        "deemed likely pathogenic, 50-75th percentile as possibly "
        "pathogenic; 25-50th percentile as possibly benign; "
        "bottom 25th percentile as likely benign. "
        "Variants reported as Confirmed pathogenic in MITOMAP have additional lines " 
        "of experimental evidence which led to confirmation of pathogenicity "
        "regardless of which percentile category the variant may belong to. "
        "see "
        "https://mitomap.org/foswiki/bin/view/MITOMAP/MitoTipInfo")
    documentation.append(
        "anticodon: TRUE if the variant falls in an anticodon of a tRNA.")
    documentation.append(
        "Columns suffix with _mitomap are annotations drawn from the MITOMAP "
        "compendium of polymorphisms and mutations in human mitochondrial DNA. "
        "see http://www.mitomap.org/MITOMAP. "
        "Variants from the the 4 tables are included: 'Control Region Variants "
        "(16024-576)', 'Coding & RNA Variants (577-16023, MTTF-MTTP)', 'rRNA/tRNA "
        " Mutations' and 'Coding & Control Region Mutations' (obtained in 2018)")
    documentation.append(
        "disease_mitomap: The disease annotation for a particular variant, "
        "as annotated by MITOMAP")
    documentation.append(
        "status_mitomap: an indication of the strength of evidence supporting "
        "the MITOMAP annotation. The strongest evidence is Confirmed, which "
        "MITMAP calls Cfrm. "
        "See https://mitomap.org/foswiki/bin/view/MITOMAP/MutationsRNACfrm")
    documentation.append(
        "GenBank_frequency_mitomap: the frequency of GenBank records "
        "containing this variant. This is a proxy for the population "
        "frequency of the variant. "
        "As expected, none of the confirmed pathogenic variants in MITOMAP "
        "have GenBank frequency >0.0%")
    documentation.append(
        "allele_frequency_mitomap: MITOMAP derives this from 32,059 GenBank "
        "full length sequences, with size greater than 15.4kbp (range 0-1)")
    documentation.append(
        "Columns suffix with _VEP come from Ensembl’s Variant Effect "
        "Predictor (VEP) annotations in the VCF")
    documentation.append(
        "highest_vep_impact: Highest VEP impact from all the VEP annotations "
        "for the variant. See http://www.ensembl.org/info/docs/tools/vep/index.html")
    documentation.append(
        "Entries separated with a semicolon mean that there are multiple "
        "annotations, from either VEP or MITOMAP.")
    documentation.append(
        "Fields from VCF header: CHR, POS, REF, ALT, QUAL, FILTER, "
        "INFO, FORMAT")
    documentation.append(
        "Additional fields from mity VCF INFO field: MQM_INFO, "
        "MQMR_INFO, QA_INFO, QR_INFO, SAF_INFO, SAR_INFO, SRF_INFO, "
        "SRR_INFO, SBR_INFO, SBA_INFO")
    documentation.append(
        "Additional fields from the mity VCF FILTER field: POS_FILTER, "
        "SBR_FILTER, SBA_FILTER, MQMR_FILTER, AQR_FILTER")
    documentation.append(
        "Additional fields from the mity VCF FORMAT field: GT_FORMAT, "
        "QR_FORMAT, AQR_FORMAT, QA_FORMAT, AQA_FORMAT")
    if dupe_header_warning:
        documentation.append(
            "Warning: VCF header IDs are not unique, so there might be a "
            "repeated definition below.")
    
    for line in noDupes_vcf_documentation:
        documentation.append(line)
    
    documentation_df = pandas.DataFrame(documentation)
    
    ## change the columns that are numerical data types to numerical
    annotated_variants1['variant_heteroplasmy'] = annotated_variants1[
        'variant_heteroplasmy'].astype('float64')
    annotated_variants1['variant_quality'] = annotated_variants1[
        'variant_quality'].astype('float64')
    annotated_variants1['ref_depth'] = annotated_variants1['ref_depth'].astype(
        'int64')
    annotated_variants1['alt_depth'] = annotated_variants1['alt_depth'].astype(
        'int64')
    annotated_variants1['total_sample_depth'] = annotated_variants1[
        'total_sample_depth'].astype('int64')
    annotated_variants1['total_locus_depth'] = annotated_variants1[
        'total_locus_depth'].astype('int64')
    annotated_variants1['COHORT_COUNT'] = annotated_variants1[
        'COHORT_COUNT'].astype('int64')
    annotated_variants1['tier'] = annotated_variants1['tier'].astype('int64')
    annotated_variants1['MQM_INFO'] = annotated_variants1['MQM_INFO'].astype(
        'float64')
    annotated_variants1['MQMR_INFO'] = annotated_variants1['MQMR_INFO'].astype(
        'float64')
    annotated_variants1['QA_INFO'] = annotated_variants1['QA_INFO'].astype(
        'float64')
    annotated_variants1['QR_INFO'] = annotated_variants1['QR_INFO'].astype(
        'float64')
    annotated_variants1['SAF_INFO'] = annotated_variants1['SAF_INFO'].astype(
        'float64')
    annotated_variants1['SAR_INFO'] = annotated_variants1['SAR_INFO'].astype(
        'float64')
    annotated_variants1['SRF_INFO'] = annotated_variants1['SRF_INFO'].astype(
        'float64')
    annotated_variants1['SRR_INFO'] = annotated_variants1['SRR_INFO'].astype(
        'float64')
    annotated_variants1['SBR_INFO'] = annotated_variants1['SBR_INFO'].astype(
        'float64')
    annotated_variants1['SBA_INFO'] = annotated_variants1['SBA_INFO'].astype(
        'float64')

    if not os.path.exists(out_folder_path):
        os.makedirs(out_folder_path)

    xlsx_name = os.path.join(out_folder_path, prefix + ".annotated_variants.xlsx")
    logging.info("saving xlsx report: " + xlsx_name)
    writer = pandas.ExcelWriter(xlsx_name, engine='xlsxwriter')
    documentation_df.to_excel(writer, sheet_name='Documentation', index=False,
                              header=False)
    annotated_variants1.to_excel(writer, sheet_name='Variants', index=False)
    writer.save()
    
    csv_name = os.path.join(out_folder_path, prefix + ".annotated_variants.csv")
    logging.info("saving csv report: " + csv_name)
    annotated_variants1.to_csv(csv_name, index=False, )
    # numpy.savetxt('data.csv', delimiter=',', X = annotated_variants1)
    # with open("test.csv", "w", newline='') as csv_file:
    #   writer = csv.writer(csv_file, delimiter=',')
    #   for line in annotated_variants1:
    #       writer.writerow(line)
